import os

import appdirs
import yaml


class PdtoolsConfig(object):
    def __init__(self, data=None):
        if data is not None:
            self.data = data
        else:
            self.data = {}

    def addAccessToken(self, netloc, username, token):
        """
        Add a new auth token for a controller or device.
        """
        if 'tokens' not in self.data:
            self.data['tokens'] = []

        self.data['tokens'].append({
            'netloc': netloc,
            'token': token,
            'username': username
        })

    def getAccessToken(self, netloc):
        """
        Get a saved auth token for a controller or device.
        """
        for entity in self.data.get('tokens', []):
            if entity['netloc'] == netloc:
                return entity.get('token', None)
        return None

    def removeAccessToken(self, token):
        """
        Remove a saved auth token for a controller or device.
        """
        old = self.data.get('tokens', [])
        self.data['tokens'] = [x for x in old if not x['token'] == token]

    def save(self):
        """
        Save the configuration file.
        """
        config_dir = appdirs.user_config_dir('pdtools', 'paradrop')
        path = os.path.join(config_dir, 'config.yaml')

        if not os.path.exists(config_dir):
            os.makedirs(config_dir)

        with open(path, 'w') as output:
            yaml.safe_dump(self.data, output, default_flow_style=False)

    @classmethod
    def load(cls):
        """
        Load the configuration file.

        Returns an empty PdconfConfig object if the file could not be read.
        """
        config_dir = appdirs.user_config_dir('pdtools', 'paradrop')
        path = os.path.join(config_dir, 'config.yaml')

        try:
            with open(path, 'r') as source:
                config = yaml.safe_load(source)
                return cls(config)
        except:
            return cls()



import re

from six.moves import input


SUPPORTED_BASE_IMAGES = set([
    "go",
    "gradle",
    "maven",
    "node",
    "python2"
])


def get_name():
    print("")
    print("Chute Name")
    print("")
    print("This name will be used on the router status page and chute store.")
    print("Please use only lowercase letters, numbers, and hypens.")
    print("")
    while True:
        name = input("name: ")
        if len(name) == 0:
            continue

        match = re.match("[a-z][a-z0-9\-]*", name)
        if match is None:
            print("The name does not meet Paradrop's requirements for chute names.")
            print("Please use only lowercase letters, numbers, and hypens.")
            continue

        return name


def get_description():
    print("")
    print("Chute Description")
    print("")
    print("Enter a description for the chute. This will be displayed in")
    print("the chute store.")
    print("")
    while True:
        desc = input("description: ")
        if len(desc) == 0:
            continue

        return desc


def get_chute_type():
    print("")
    print("Chute Type")
    print("")
    print("Paradrop has two types of chutes. Light chutes are based on a base")
    print("image that is optimized for a particular language such as Python or")
    print("JavaScript and use the language-specific package manager (pip or npm)")
    print("to install dependencies. Normal chutes give you more flexibility to")
    print("install dependencies but require that you write your own Dockerfile.")
    print("")
    print("Valid types: light, normal")
    valid = set(["light", "normal"])
    while True:
        ctype = input("type [normal]: ").lower()
        if len(ctype) == 0:
            ctype = "normal"

        if ctype not in valid:
            print("Valid types: light, normal")
            continue

        return ctype


def get_base_name():
    print("")
    print("Base Image Name")
    print("")
    print("Enter the name of the base image to use. This depends on the")
    print("programming language that you intend to use.")
    print("")
    print("Valid choices: " + ", ".join(SUPPORTED_BASE_IMAGES))
    while True:
        name = input("image [python2]: ").lower()
        if len(name) == 0:
            name = "python2"

        if name not in SUPPORTED_BASE_IMAGES:
            print("Valid choices: " + ", ".join(SUPPORTED_BASE_IMAGES))
            continue

        return name


def get_command(name="my-app", use=None):
    print("")
    print("Command")
    print("")
    print("This is the command to start your application.")

    default = None
    if use == "go":
        print("For Go applications, it is highly recommended that you use the default.")
        default = "app"
    elif use == "gradle":
        print("For Java applications, you can accept the default and update")
        print("paradrop.yaml after you have decided on the structure of your project.")
        default = "java -cp build/libs/gradle-{name}-1.0-SNAPSHOT.jar com.mycompany.{name}.Main".format(name=name)
    elif use == "maven":
        print("For Java applications, you can accept the default and update")
        print("paradrop.yaml after you have decided on the structure of your project.")
        default = "java -cp target/{name}-1.0-SNAPSHOT.jar com.mycompany.{name}.Main".format(name=name)
    elif use == "node":
        default = "node index.js"
    elif use == "python2":
        default = "python2 -u main.py"
    else:
        default = ""
    print("")

    entry = input("command [{}]: ".format(default))
    if len(entry) == 0:
        entry = default

    return entry


def build_chute():
    chute = {
        'version': 1,
        'config': {}
    }

    chute['name'] = get_name()
    chute['description'] = get_description()
    chute['type'] = get_chute_type()

    if chute['type'] == "light":
        chute['use'] = get_base_name()
        chute['command'] = get_command(name=chute['name'], use=chute['use'])

    return chute



"""
Paradrop command line utility.

Two guiding principles were used in the design of the Paradrop command
line utility.

* The command tree should be relatively flat for easy discoverability.
Commands are grouped into a few cohesive modules (chute, cloud, node,
and store).  All of the commands in the cloud module make use of the
cloud controller API, for example. There are no further subgroups under
these top-level groups, which means that `pdtools cloud --help` lists
all relevant commands.

* Commands should be self-evident in their effect. All command names
begin with a verb. Commands that begin with "list-" can be expected
to produce a table of output; commands that begin with "create-" can
be expected to result in the creation of a new resource; and so on.
Although "describe-" is a bit verbose, it conveys a sense that this
command returns everything one needs to know about an object in a way that
"get" does not. It is possible to enable tab completion when pdtools
is installed as a system package.

To enable tab completion, add the following line to your .bashrc file:

    eval "$(_PDTOOLS_COMPLETE=source pdtools)"
"""
import os

import click

# Deprecated command groups to be removed prior to 1.0 release
from . import device
from . import groups
from . import routers

# Command groups
from . import chute
from . import cloud
from . import node
from . import store
from . import wizard


PDSERVER_URL = os.environ.get("PDSERVER_URL", "https://paradrop.org")

CONTEXT_SETTINGS = dict(
    # Options can be parsed from PDTOOLS_* environment variables.
    auto_envvar_prefix = 'PDTOOLS',

    # Respond to both -h and --help for all commands.
    help_option_names = ['-h', '--help'],

    obj = {
        'pdserver_url': PDSERVER_URL
    }
)


@click.group(context_settings=CONTEXT_SETTINGS)
def root():
    """
    Paradrop command line utility.
    """
    pass


@root.command('help')
@click.pass_context
def help(ctx):
    """
    Show this message and exit
    """
    click.echo(ctx.parent.get_help())


# Deprecated command groups to be removed prior to 1.0 release
root.add_command(device.device)
root.add_command(routers.routers)
groups.register_commands(root)

# Command groups
root.add_command(chute.root)
root.add_command(cloud.root)
root.add_command(node.root)
root.add_command(store.root)
root.add_command(wizard.root)


def main():
    """
    Entry point for the pdtools Python package.
    """
    root()


if __name__ == "__main__":
    main()



#
# This module is DEPRECATED and should be removed prior to the 1.0 release.
#
# Most functionality has been moved to the AuthenticatedClient base class along
# with the ParadropClient and ControllerClient child classes. Some of the old
# command trees (device, group, routers) still depend on the functions in this
# module, and in particular, we should keep the "device" command tree around
# until the 1.0 release because the router_request function here supports
# Paradrop versions prior to 0.10.
#

from __future__ import print_function

import base64
import os
from pprint import pprint
from six.moves.urllib.parse import urlparse

import requests
import websocket

from .config import PdtoolsConfig
from .util import LoginGatherer


PDSERVER_URL = os.environ.get('PDSERVER_URL', 'https://paradrop.org')

LOCAL_DEFAULT_USERNAME = "paradrop"
LOCAL_DEFAULT_PASSWORD = ""


def change_json(data, path, value):
    """
    Replace a value in a JSON object.
    """
    parts = path.split('.')
    head = data
    for i in range(len(parts)-1):
        if isinstance(head, list):
            parts[i] = int(parts[i])
        head = head[parts[i]]
    head[parts[-1]] = value


def pdserver_request(method, url, json=None, headers=None):
    """
    Issue a Paradrop controller API request.

    This will prompt for a username and password if necessary.
    """
    session = requests.Session()

    # Extract just the hostname from the controller URL.  This will be the
    # authentication domain.
    parts = urlparse(PDSERVER_URL)

    config = PdtoolsConfig.load()
    token = config.getAccessToken(parts.netloc)

    while True:
        while token is None:
            username, password = LoginGatherer.prompt(parts.netloc)
            data = {
                'email': username,
                'password': password
            }

            auth_url = '{}/auth/local'.format(PDSERVER_URL)
            request = requests.Request('POST', auth_url, json=data, headers=headers)
            prepped = session.prepare_request(request)
            res = session.send(prepped)
            print("Server responded: {} {}".format(res.status_code, res.reason))

            try:
                res_data = res.json()
                token = res_data['token']

                config.addAccessToken(parts.netloc, username, token)
                config.save()
            except:
                pass

        session.headers.update({'Authorization': 'Bearer {}'.format(token)})
        request = requests.Request(method, url, json=json, headers=headers)
        prepped = session.prepare_request(request)

        res = session.send(prepped)
        print("Server responded: {} {}".format(res.status_code, res.reason))
        if res.status_code == 401:
            # Token is probably expired.
            config.removeAccessToken(token)
            config.save()
            token = None
            continue
        else:
            return res


def router_login(base_url):
    """
    Prompt for router username and password and attempt login.

    Returns the username if successful or None.
    """
    config = PdtoolsConfig.load()
    session = requests.Session()
    url_parts = urlparse(base_url)

    for username, password in LoginGatherer(url_parts.netloc):
        # Try to get a token for later use. Prior to 1.10, paradrop-daemon
        # does not not support tokens.
        _, token = send_router_login(base_url, username, password, session)
        if token is not None:
            config.addAccessToken(url_parts.netloc, username, token)
            config.save()
            return username

    return None


def router_logout(base_url):
    """
    Delete access token(s) for device if they exist.

    Returns the number of tokens removed.
    """
    config = PdtoolsConfig.load()
    url_parts = urlparse(base_url)

    # Generally, there should be only zero or one access token, but it is
    # possible for a code bug or other situation to cause there to be multiple
    # tokens.  The loop makes sure we remove them all.
    removed = 0
    while True:
        token = config.getAccessToken(url_parts.netloc)
        if token is None:
            break
        else:
            config.removeAccessToken(token)
            config.save()
            removed += 1

    return removed


def send_router_login(request_url, username, password, session):
    url_parts = urlparse(request_url)

    auth_url = "{}://{}/api/v1/auth/local".format(url_parts.scheme, url_parts.netloc)
    data = {
        'username': username,
        'password': password
    }
    request = requests.Request('POST', auth_url, json=data)
    prepped = session.prepare_request(request)
    res = session.send(prepped)
    print("Server responded: {} {}".format(res.status_code, res.reason))

    if res.ok:
        data = res.json()
        token = data['token']

        return (res.status_code, token)

    else:
        return (res.status_code, None)


def router_request(method, url, json=None, headers=None, dump=True, data=None, **kwargs):
    """
    Issue a router API request.

    This will prompt for a username and password if necessary.
    """
    session = requests.Session()
    request = requests.Request(method, url, json=json, headers=headers,
            data=data, **kwargs)

    # Extract just the hostname from the controller URL.  This will be the
    # authentication domain.
    url_parts = urlparse(url)

    config = PdtoolsConfig.load()
    token = config.getAccessToken(url_parts.netloc)

    # If data is a file or file-like stream, then every time we read from it,
    # the read pointer moves to the end, and we will need to reset it if we are
    # to read again.
    try:
        pos = data.tell()
    except AttributeError:
        pos = None

    if token is not None:
        session.headers.update({'Authorization': 'Bearer {}'.format(token)})
        prepped = session.prepare_request(request)

        res = session.send(prepped)
        print("Server responded: {} {}".format(res.status_code, res.reason))
        if res.status_code == 401:
            # Token has probably expired.
            config.removeAccessToken(token)
            config.save()
            token = None
        else:
            # Not an auth error, so return the result.
            if res.ok and dump:
                data = res.json()
                pprint(data)
            return res

    for username, password in LoginGatherer(url_parts.netloc):
        # Try to get a token for later use. Prior to 0.10, paradrop-daemon
        # does not not support tokens.
        _, token = send_router_login(url, username, password, session)
        if token is not None:
            config.addAccessToken(url_parts.netloc, username, token)
            config.save()

        userpass = "{}:{}".format(username, password).encode('utf-8')
        encoded = base64.b64encode(userpass.encode('utf-8')).decode('ascii')
        session.headers.update({'Authorization': 'Basic {}'.format(encoded)})
        prepped = session.prepare_request(request)

        # Reset the read pointer if the body comes from a file.
        if pos is not None:
            prepped.body.seek(pos)

        res = session.send(prepped)
        print("Server responded: {} {}".format(res.status_code, res.reason))
        if res.status_code == 401:
            # Probably incorrect username/password, so try again.
            continue
        else:
            if res.ok and dump:
                data = res.json()
                pprint(data)
            return res


def router_ws_request(url, headers=None, **kwargs):
    """
    Issue a router WS connection.

    This will prompt for a username and password if necessary.
    """
    if headers is None:
        headers = {}
    else:
        headers = headers.copy()

    def on_error(ws, x):
        print(x)

    url_parts = urlparse(url)
    config = PdtoolsConfig.load()
    token = config.getAccessToken(url_parts.netloc)

    if token is not None:
        headers.update({'Authorization': 'Bearer {}'.format(token)})
    else:
        username, password = LoginGatherer.prompt(url_parts.netloc)

        userpass = "{}:{}".format(username, password)
        encoded = base64.b64encode(userpass.encode('utf-8')).decode('ascii')
        headers.update({'Authorization': 'Basic {}'.format(encoded)})

    ws = websocket.WebSocketApp(url, header=headers, **kwargs)
    ws.on_error = on_error
    ws.run_forever()



"""
The TokenProvider abstraction allows for different methods of acquiring a token
for authorization with an HTTP service. As shown in the table below, multiple
methods of acquiring a token are generally available, but their applicability
depends on whether the code is running interactively or inside a chute where
there is no user input and whether the target API is the node or the cloud
controller.  This creates a fair bit of complexity for the AuthenticatedClient
to perform an authenticated API call.

Method               | Caller    | Target API
(x = applicable)     | CLI Chute | Node Cloud
---------------------------------------------
Environment variable | x   x     | x    -
Saved locally        | x   -     | x    x
Default password     | x   -     | x    -
Prompt for input     | x   -     | x    x

Here is an example flow for the more complicated case, interacting with a
node's API:

1. We check for an environment variable because chutes authenticate using
   a token provided through the PARADROP_API_TOKEN variable.
2. We check for a token saved to a file. We save tokens locally so that
   users do not need to type their password with each command.
3. If there is not a saved token or if the saved token has expired, then
   we can try a default password. Most nodes intended for development
   or education purposes retain their default settings.
4. Finally, if user input is available, we can prompt the user to enter
   a username and password. We should generally never prompt for a
   password in code that runs inside a chute. That code is presumed to
   be non-interactive, and it would be better just to return an error.

The TokenProvider abstraction helps simplify the logic above.  It allows an
AuthenticatedClient implementation to instantiate one or more parameterized
TokenProvider objects. With a list of provider objects, the sequence of steps
above amounts to iterating the list and calling `get_token` on each provider
until one of them succeeds.
"""

import getpass
import os
import sys

from six.moves.urllib.parse import urlparse

import builtins
import requests

from .config import PdtoolsConfig


class TokenProvider(object):
    """
    Abstract class for methods that provide authentication tokens.
    """
    def get_token(self):
        return NotImplemented

    def is_applicable(self):
        return False

    def should_retry(self):
        return False

    def update(self, token, success):
        pass


class EnvironmentVariableTokenProvider(TokenProvider):
    def __init__(self, variable="PARADROP_API_TOKEN"):
        self.variable = variable

    def get_token(self):
        return os.environ.get(self.variable, None)

    def is_applicable(self):
        return self.variable in os.environ


class SavedTokenProvider(TokenProvider):
    def __init__(self, domain):
        self.domain = domain

    def get_token(self):
        config = PdtoolsConfig.load()
        return config.getAccessToken(self.domain)

    def is_applicable(self):
        return True

    def update(self, token, success):
        if not success:
            config = PdtoolsConfig.load()
            config.removeAccessToken(token)
            config.save()


class DefaultLoginTokenProvider(TokenProvider):
    def __init__(self, auth_url, param_map):
        self.auth_url = auth_url
        self.param_map = param_map

    def get_token(self):
        url_parts = urlparse(self.auth_url)

        data = {
            self.param_map['username']: 'paradrop',
            self.param_map['password']: ''
        }
        res = requests.post(self.auth_url, json=data)
        try:
            data = res.json()
            return data['token']
        except:
            return None

    def is_applicable(self):
        return True


class LoginPromptTokenProvider(TokenProvider):
    def __init__(self, auth_url, param_map):
        self.auth_url = auth_url
        self.param_map = param_map

        self.username = None
        self.token = None

    def get_token(self):
        url_parts = urlparse(self.auth_url)

        print("Attempting to log in to authentication domain {}".format(url_parts.netloc))
        self.username = builtins.input("Username: ")
        password = getpass.getpass("Password: ")

        data = {
            self.param_map['username']: self.username,
            self.param_map['password']: password
        }
        res = requests.post(self.auth_url, json=data)
        try:
            data = res.json()
            self.token = data['token']
            return self.token
        except:
            return None

    def is_applicable(self):
        # Not sure if this works. The intent is to prompt for a password if
        # called from the command line but skip this method if running inside a
        # chute.
        return sys.stdin.isatty()

    def update(self, token, success):
        if success and token == self.token:
            url_parts = urlparse(self.auth_url)
            config = PdtoolsConfig.load()
            config.addAccessToken(url_parts.netloc, self.username, token)
            config.save()



import builtins
import click
import json
import operator
import os

import git
import yaml

from .controller_client import ControllerClient
from .helpers.chute import build_chute
from .util import update_object


def chute_find_field(chute, key, default=Exception):
    """
    Find a field in a chute definition loading from a paradrop.yaml file.
    """
    if key in chute:
        return chute[key]
    elif 'config' in chute and key in chute['config']:
        return chute['config'][key]
    elif isinstance(default, type):
        raise default("{} field not found in chute definition.".format(key))
    else:
        return default


def chute_resolve_source(source, config):
    """
    Resolve the source section from paradrop.yaml to store configuration.

    If git/http, add an appropriate download section to the chute
    configuration. For git repos, we also identify the latest commit and add
    that to the download information. If type is inline, add a dockerfile
    string field to the chute and no download section.
    """
    if 'type' not in source:
        raise Exception("Source type not specified for chute.")

    source_type = source['type']
    if source_type == 'http':
        config['download'] = {
            'url': source['url']
        }

    elif source_type == 'git':
        repo = git.Repo('.')
        config['download'] = {
            'url': source['url'],
            'checkout': str(repo.head.commit)
        }

    elif source_type == 'inline':
        with open('Dockerfile', 'r') as dockerfile:
            config['dockerfile'] = dockerfile.read()

    else:
        raise Exception("Invalid source type {}".format(source_type))


@click.group('chute')
@click.pass_context
def root(ctx):
    """
    Utilities for developing chutes.

    These commands all operate on a chute project in the current working
    directory.  Remember, that all chutes must have a paradrop.yaml file
    in the top-level directory.  You can create one interactively with
    the initialize command.
    """
    pass


@root.command('add-wifi-ap')
@click.argument('essid')
@click.option('--password', default=None,
        help='Password for the network, must be at least 8 characters if specified.')
@click.option('--force', default=False, is_flag=True,
        help='Overwrite an existing section in the configuration file.')
@click.pass_context
def add_wifi_ap(ctx, essid, password, force):
    """
    Add a WiFi AP to the chute configuration.

    ESSID must be between 1 and 32 characters in length.  Spaces are
    allowed if you enclose the argument in quotation marks.
    """
    with open('paradrop.yaml', 'r') as source:
        chute = yaml.safe_load(source)

    # Make sure the config.net section exists.
    net = update_object(chute, 'config.net')

    # Only support adding the first Wi-Fi interface for now.
    if 'wifi' in net and not force:
        click.echo("Wi-Fi interface already exists in paradrop.yaml.")
        click.echo("Please edit the file directly to make changes.")
        return

    net['wifi'] = {
        'type': 'wifi',
        'intfName': 'wlan0',
        'dhcp': {
            'lease': '12h',
            'start': 4,
            'limit': 250
        },
        'ssid': essid,
        'options': {
            'isolate': False,
            'hidden': False
        }
    }

    # Minimum password length is part of the standard.
    if password is not None:
        if len(password) < 8:
            print("Wi-Fi password must be at least 8 characters.")
            return
        net['wifi']['key'] = password

    with open('paradrop.yaml', 'w') as output:
        yaml.safe_dump(chute, output, default_flow_style=False)


@root.command('export-configuration')
@click.option('--format', '-f', default='json',
        help="Format (json or yaml)")
@click.pass_context
def export_configuration(ctx, format):
    """
    Export chute configuration in JSON or YAML format.

    The configuration format used by the cloud API is slightly different
    from the paradrop.yaml file. This command can export a JSON object in
    a form suitable for installing the chute through the cloud API.

    The config object will usually be used in an envelope like the following:
    {
      "updateClass": "CHUTE",
      "updateType": "update",
      "config": { config-object }
    }
    """
    with open('paradrop.yaml', 'r') as source:
        chute = yaml.safe_load(source)

    if 'source' not in chute:
        click.echo("Error: source section missing in paradrop.yaml")
        return

    config = chute.get('config', {})

    config['name'] = chute['name']
    config['version'] = chute.get('version', 1)

    chute_resolve_source(chute['source'], config)

    format = format.lower()
    if format == "json":
        click.echo(json.dumps(config, sort_keys=True, indent=2))
    elif format == "yaml":
        click.echo(yaml.safe_dump(config, default_flow_style=False))
    else:
        click.echo("Unrecognized format: {}".format(format))


@root.command('initialize')
@click.pass_context
def initialize(ctx):
    """
    Interactively create a paradrop.yaml file.
    """
    chute = build_chute()
    with open("paradrop.yaml", "w") as output:
        yaml.safe_dump(chute, output, default_flow_style=False)

    # If this is a node.js chute, generate a package.json file from the
    # information that the user provided.
    if chute.get('use', None) == 'node':
        if not os.path.isfile('package.json'):
            data = {
                'name': chute['name'],
                'version': '1.0.0',
                'description': chute['description']
            }
            with open('package.json', 'w') as output:
                json.dump(data, output, sort_keys=True, indent=2)


@root.command('set')
@click.argument('path')
@click.argument('value')
@click.pass_context
def set_config(ctx, path, value):
    """
    Set a value in the paradrop.yaml file.

    PATH must be a dot-separated path to a value in the paradrop.yaml
    file, such as "config.web.port". VALUE will be interpreted as a
    string, numeric, or boolean type as appropriate.

    Changing values inside a list is not currently supported.  For that
    you will need to edit the file directly.

    Example: set config.web.port 80
    """
    with open('paradrop.yaml', 'r') as source:
        chute = yaml.safe_load(source)

    # Empty paradrop.yaml file?
    if chute is None:
        chute = {}

    # Calling yaml.safe_load on the value does a little type interpretation
    # (e.g. numeric vs. string) and allows the user to directly set lists and
    # other structures.
    value = yaml.safe_load(value)

    def set_value(parent, key, created):
        if created:
            click.echo("Creating new field {} = {}".format(path, value))
        else:
            current = parent[key]
            click.echo("Changing {} from {} to {}".format(path, current, value))

        parent[key] = value

    update_object(chute, path, set_value)

    with open('paradrop.yaml', 'w') as output:
        yaml.safe_dump(chute, output, default_flow_style=False)



import os
import re

from .authenticated_client import AuthenticatedClient
from .devices.camera import Camera


LOCAL_DEFAULT_USERNAME = "paradrop"
LOCAL_DEFAULT_PASSWORD = ""

PARADROP_API_TOKEN = os.environ.get("PARADROP_API_TOKEN", None)
PARADROP_CHUTE_NAME = os.environ.get("PARADROP_CHUTE_NAME", None)


class ParadropClient(AuthenticatedClient):
    """
    Client for interacting with a Paradrop daemon instance.

    Here is a motivating example that is intended to be valid chute code:

    from pdtools import ParadropClient
    client = ParadropClient()
    for camera in client.get_cameras():
        img = camera.get_image()
    """
    def __init__(self, host=None):
        if host is None:
            host = "172.17.0.1"

        super(ParadropClient, self).__init__("node", "http://"+host)

        self.host = host
        self.base_url = "http://{}/api/v1".format(host)
        self.snapd_url = "http://{}/snapd/v2".format(host)


    def add_change(self, change):
        """
        Schedule a new change on the node.
        """
        url = "{}/changes/".format(self.base_url)
        return self.request("POST", url, json=change)

    def add_ssh_key(self, key_text, user="paradrop"):
        """
        Add an authorized key for SSH access.
        """
        url = "{}/config/sshKeys/{}".format(self.base_url, user)
        data = {
            'key': key_text
        }
        return self.request("POST", url, json=data)

    def connect_snap_interface(self, slots=[], plugs=[]):
        """
        Connect an interface for an installed snap.
        """
        url = self.snapd_url + "/interfaces"
        data = {
            "action": "connect",
            "slots": slots,
            "plugs": plugs
        }
        return self.request("POST", url, json=data)

    def create_user(self, email):
        """
        Create a local user connected to an Ubuntu One account.
        """
        url = self.snapd_url + "/create-user"
        data = {
            "email": email,
            "sudoer": True
        }
        return self.request("POST", url, json=data)

    def generate_config(self):
        """
        Generate a new node configuration based on detected hardware.
        """
        url = self.base_url + "/config/new-config"
        return self.request("GET", url)

    def get_audio(self):
        """
        Get information about the audio subsystem.
        """
        url = self.base_url + "/audio/info"
        return self.request("GET", url)

    def get_chute(self, chute_name):
        """
        Get information about an installed chute.
        """
        url = self.base_url + "/chutes/" + chute_name
        return self.request("GET", url)

    def get_chute_cache(self, chute_name):
        """
        Get internal information about an installed chute.
        """
        url = "{}/chutes/{}/cache".format(self.base_url, chute_name)
        return self.request("GET", url)

    def get_chute_client(self, chute_name, network, client):
        """
        Get information about a chute's network client.
        """
        url = "{}/chutes/{}/networks/{}/stations/{}".format(self.base_url,
                chute_name, network, client)
        return self.request("GET", url)

    def get_chute_config(self, chute_name):
        """
        Get chute configuration.
        """
        url = "{}/chutes/{}/config".format(self.base_url, chute_name)
        return self.request("GET", url)

    def get_config(self):
        """
        Get node configuration.
        """
        url = self.base_url + "/config/hostconfig"
        return self.request("GET", url)

    def get_pdconf(self):
        """
        Get status of the pdconf subsystem.
        """
        url = self.base_url + "/config/pdconf"
        return self.request("GET", url)

    def get_provision(self):
        """
        Get provisioning status of the node.
        """
        url = self.base_url + "/config/provision"
        return self.request("GET", url)

    def install_tar(self, data, name=None):
        """
        Install a chute from a tar file.
        """
        url = self.base_url + "/chutes/" # trailing slash is intentional

        if name is None:
            method = "POST"
        else:
            method = "PUT"
            url = url + name

        headers = {'Content-Type': 'application/x-tar'}
        return self.request(method, url, headers=headers, data=data)

    def list_audio_modules(self):
        """
        List modules loaded by the audio subsystem.
        """
        url = self.base_url + "/audio/modules"
        return self.request("GET", url)

    def list_audio_sinks(self):
        """
        List audio sinks.
        """
        url = self.base_url + "/audio/sinks"
        return self.request("GET", url)

    def list_audio_sources(self):
        """
        List audio sources.
        """
        url = self.base_url + "/audio/sources"
        return self.request("GET", url)

    def list_changes(self):
        """
        List queued or in-progress changes.
        """
        url = self.base_url + "/changes/"
        return self.request("GET", url)

    def list_chute_clients(self, chute_name, network):
        """
        List clients connected to the chute's network
        """
        url = "{}/chutes/{}/networks/{}/stations".format(self.base_url,
                chute_name, network)
        return self.request("GET", url)

    def list_chute_networks(self, chute_name):
        """
        List networks configured by the chute.
        """
        url = "{}/chutes/{}/networks".format(self.base_url, chute_name)
        return self.request("GET", url)

    def list_chutes(self):
        """
        List chutes installed on the node.
        """
        url = self.base_url + "/chutes/" # trailing slash is intentional
        return self.request("GET", url)

    def list_snap_interfaces(self):
        """
        List interfaces for snaps installed on the node.
        """
        url = self.snapd_url + "/interfaces"
        return self.request("GET", url)

    def list_ssh_keys(self, user="paradrop"):
        """
        List authorized keys for SSH access.
        """
        url = "{}/config/sshKeys/{}".format(self.base_url, user)
        return self.request("GET", url)

    def load_audio_module(self, module_name):
        """
        Load a module into the audio subsystem.
        """
        url = "{}/audio/modules".format(self.base_url)
        data = {
            "name": module_name
        }
        return self.request("POST", url, json=data)

    def provision(self, id, key, controller=None, wamp=None):
        """
        Provision the node by connecting to a cloud controller.
        """
        url = self.base_url + "/config/provision"

        data = {
            'routerId': id,
            'apitoken': key
        }
        if controller is not None:
            data['pdserver'] = controller
        if wamp is not None:
            data['wampRouter'] = wamp

        return self.request("POST", url, json=data)

    def remove_chute(self, chute_name):
        """
        Remove a chute from the node.
        """
        url = self.base_url + "/chutes/" + chute_name
        return self.request("DELETE", url)

    def remove_chute_client(self, chute_name, network, client):
        """
        Remove a connected client from the chute's network.
        """
        url = "{}/chutes/{}/networks/{}/stations/{}".format(self.base_url,
                chute_name, network, client)
        return self.request("DELETE", url)

    def restart_chute(self, chute_name):
        """
        Restart a chute.
        """
        url = "{}/chutes/{}/restart".format(self.base_url, chute_name)
        return self.request("POST", url)

    def set_chute_config(self, chute_name, config):
        """
        Set chute configuration.
        """
        url = "{}/chutes/{}/config".format(self.base_url, chute_name)
        return self.request("PUT", url, json=config)

    def set_chute_variables(self, chute_name, config):
        """
        Set chute environment variables.
        """
        url = "{}/chutes/{}/restart".format(self.base_url, chute_name)
        data = {
            "environment": config
        }
        return self.request("POST", url, json=data)

    def set_config(self, config):
        """
        Set node configuration.
        """
        url = self.base_url + "/config/hostconfig"
        data = {
            'config': config
        }
        return self.request("PUT", url, json=data)

    def set_password(self, username, password):
        """
        Set the local admin password.
        """
        url = self.base_url + "/password/change"
        data = {
            "username": username,
            "password": password
        }
        return self.request("POST", url, json=data)

    def set_sink_volume(self, sink, volumes):
        """
        Set audio sink volume.
        """
        url = "{}/audio/sinks/{}/volume".format(self.base_url, sink)
        return self.request("PUT", url, json=volumes)

    def set_source_volume(self, source, volumes):
        """
        Set audio source volume.
        """
        url = "{}/audio/sources/{}/volume".format(self.base_url, source)
        return self.request("PUT", url, json=volumes)

    def start_chute(self, chute_name):
        """
        Start a stopped chute.
        """
        url = "{}/chutes/{}/start".format(self.base_url, chute_name)
        return self.request("POST", url)

    def stop_chute(self, chute_name):
        """
        Stop a running chute.
        """
        url = "{}/chutes/{}/stop".format(self.base_url, chute_name)
        return self.request("POST", url)

    def trigger_pdconf(self):
        """
        Trigger pdconf to reload configuration.
        """
        url = self.base_url + "/config/pdconf"
        return self.request("PUT", url)



    def get_cameras(self, chute_name=PARADROP_CHUTE_NAME, network_name=None):
        """
        List cameras connected to a chute's network.

        Note: This only detects the subset of D-Link cameras that we currently
        support. We will add more devices as we test them.

        chute_name: if not specified, will be read from PARADROP_CHUTE_NAME
        network_name: if not specified, will include all of the chute's networks
        """
        if chute_name is None:
            raise Exception("chute_name was not specified")

        cameras = []
        leases = self.get_leases(chute_name, network_name)

        dlink_re = re.compile("(28:10:7b|b0:c5:54|01:b0:c5):.*")
        for lease in leases:
            match = dlink_re.match(lease['mac_addr'])
            if match is not None:
                cameras.append(Camera(lease['mac_addr'], lease['ip_addr']))

        return cameras

    def get_leases(self, chute_name=PARADROP_CHUTE_NAME, network_name=None):
        """
        List DHCP lease records for a chute's network.

        chute_name: if not specified, will be read from PARADROP_CHUTE_NAME
        network_name: if not specified, will include all of the chute's networks
        """
        if chute_name is None:
            raise Exception("chute_name was not specified")

        networks = []
        devices = []

        if network_name is not None:
            networks = [network_name]
        else:
            netlist = self.get_networks(chute_name)

            # In rare cases, get_networks returns None early in the chute
            # lifetime. Instead of causing an exception, treat it as an empty
            # list - no cameras detected yet.
            if netlist is not None:
                networks = [x['name'] for x in netlist]

        for net in networks:
            url = self.base_url + "/chutes/{}/networks/{}/leases".format(
                    chute_name, net)
            result = self.request("GET", url)
            if isinstance(result, list):
                devices.extend(result)

        return devices

    def get_networks(self, chute_name=PARADROP_CHUTE_NAME):
        """
        List networks owned by a chute.

        chute_name: if not specified, will be read from PARADROP_CHUTE_NAME
        """
        if chute_name is None:
            raise Exception("chute_name was not specified")

        url = self.base_url + "/chutes/{}/networks".format(chute_name)
        result = self.request("GET", url)

        return result



from six.moves.urllib.parse import urlparse

import requests

from .config import PdtoolsConfig
from .token_provider import EnvironmentVariableTokenProvider, SavedTokenProvider, DefaultLoginTokenProvider, LoginPromptTokenProvider


class AuthenticatedClient(object):
    ApiSpecs = {
        "cloud": {
            "auth_failure_codes": set([401]),
            "auth_url": "{scheme}://{netloc}/auth/local",
            "param_map": {
                "username": "email",
                "password": "password"
            }
        },
        "node": {
            "auth_failure_codes": set([401]),
            "auth_url": "{scheme}://{netloc}/api/v1/auth/local",
            "param_map": {
                "username": "username",
                "password": "password"
            }
        }
    }

    def __init__(self, api_spec, url):
        """
        Initialize AuthenticatedClient object.

        api_spec: specifies the API in use (should be "node" or "cloud")
        """
        self.api_spec = AuthenticatedClient.ApiSpecs[api_spec]

        url_parts = urlparse(url)
        self.auth_domain = url_parts.netloc

        auth_template = self.api_spec['auth_url']
        self.auth_url = auth_template.format(scheme=url_parts.scheme,
                netloc=url_parts.netloc)

        # List of token providers. We will try these in order to obtain a valid
        # auth token in order to fulfill a request. Note that some of these are
        # parameterized based on the API spec. For example, the node and cloud
        # APIs have a different format for the authentication URL.
        self.token_providers = [
            EnvironmentVariableTokenProvider(),
            SavedTokenProvider(self.auth_domain),
            DefaultLoginTokenProvider(self.auth_url, self.api_spec['param_map']),
            LoginPromptTokenProvider(self.auth_url, self.api_spec['param_map'])
        ]

    def login(self):
        config = PdtoolsConfig.load()
        token = config.getAccessToken(self.auth_domain)
        if token is not None:
            return token

        provider = LoginPromptTokenProvider(self.auth_url, self.api_spec['param_map'])
        token = provider.get_token()
        if token is not None:
            # If the token is non-empty, we assume it must be valid.  Call
            # provider.update to make sure it gets saved.
            provider.update(token, True)
        return token

    def logout(self):
        config = PdtoolsConfig.load()

        # Generally, there should be only zero or one access token, but it is
        # possible for a code bug or other situation to cause there to be multiple
        # tokens.  The loop makes sure we remove them all.
        removed = 0
        while True:
            token = config.getAccessToken(self.auth_domain)
            if token is None:
                break
            else:
                config.removeAccessToken(token)
                config.save()
                removed += 1

        return removed

    def request(self, method, url, authenticated=True, json=None, headers=None, data=None, **kwargs):
        session = requests.Session()
        request = requests.Request(method, url, json=json, headers=headers,
                data=data, **kwargs)

        # Extract just the hostname from the controller URL.  This will be the
        # authentication domain.
        url_parts = urlparse(url)

        # If data is a file or file-like stream, then every time we read from it,
        # the read pointer moves to the end, and we will need to reset it if we are
        # to read again.
        try:
            pos = data.tell()
        except AttributeError:
            pos = None

        # Non-authenticated calls (e.g. to register an account) can bypass the
        # token provider logic.
        if not authenticated:
            prepped = session.prepare_request(request)
            res = session.send(prepped)
            try:
                return res.json()
            except:
                return None

        for provider in self.token_providers:
            if not provider.is_applicable():
                continue

            token = provider.get_token()
            session.headers.update({'Authorization': 'Bearer {}'.format(token)})
            prepped = session.prepare_request(request)

            # Reset the read pointer if the body comes from a file.
            if pos is not None:
                prepped.body.seek(pos)

            res = session.send(prepped)
            if res.status_code in self.api_spec['auth_failure_codes']:
                # The status code indicates an authorization failure.  That
                # could mean the token is expired, the password is incorrect,
                # etc.
                provider.update(token, False)
            else:
                provider.update(token, True)
                if res.status_code == 204:
                    return None
                else:
                    return res.json()



import click

from .comm import pdserver_request


@click.group()
@click.pass_context
def routers(ctx):
    """
    (deprecated) Access router information on the controller.

    These commands are deprecated. Please use the equivalent commands under
    `pdtools cloud --help`.
    """
    ctx.obj['routers_url'] = ctx.obj['pdserver_url'] + "/api/routers"


@routers.command()
@click.pass_context
def list(ctx):
    """
    List routers.
    """
    url = ctx.obj['routers_url']
    result = pdserver_request('GET', url)
    routers = result.json()

    for router in routers:
        print("{} {} {}".format(router['_id'], router['name'], router['online']))


@routers.command()
@click.pass_context
@click.argument('token')
def claim(ctx, token):
    """
    Claim an existing router.
    """
    url = ctx.obj['routers_url'] + '/claim'
    data = {
        'claim_token': token
    }
    result = pdserver_request('POST', url, json=data)
    if result.ok:
        router = result.json()
        print("Claimed router: {}".format(router['name']))
    else:
        print("There was an error claiming the router.")
        print("Please check that your claim token is correct.")


@routers.command()
@click.pass_context
@click.argument('name')
@click.option('--orphaned/--not-orphaned', default=False)
@click.option('--claim', default=None)
def create(ctx, name, orphaned, claim):
    """
    Create a new router.
    """
    url = ctx.obj['routers_url']
    data = {
        'name': name,
        'orphaned': orphaned
    }
    if claim is not None:
        data['claim_token'] = claim
    result = pdserver_request('POST', url, json=data)
    if result.ok:
        router = result.json()

        print("Name: {}".format(router['name']))
        print("ID: {}".format(router['_id']))
        print("Password: {}".format(router['password']))
        print("Claim: {}".format(router['claim_token']))

        print("")
        print("You may use the following command to provision your router:")
        print("pdtools device <address> provision {_id} {password}".format(**router))


@routers.command()
@click.pass_context
@click.argument('router_id')
def delete(ctx, router_id):
    """
    Delete a router.
    """
    url = ctx.obj['routers_url'] + "/" + router_id
    pdserver_request('DELETE', url)



import builtins
import click
import operator
import os
import yaml

from . import util
from .chute import chute_resolve_source
from .controller_client import ControllerClient


@click.group('store')
@click.pass_context
def root(ctx):
    """
    Publish and install from the public chute store.

    By default the cloud controller is assumed to be paradrop.org. This
    can be configured through the environment variable PDSERVER_URL.

    It is recommended that you log in with the cloud login command before
    using the store commands.
    """
    pass


@root.command('create-version')
@click.pass_context
def create_version(ctx):
    """
    Push a new version of the chute to the store.
    """
    if not os.path.exists("paradrop.yaml"):
        raise Exception("No paradrop.yaml file found in working directory.")

    with open('paradrop.yaml', 'r') as source:
        chute = yaml.safe_load(source)

    name = chute_find_field(chute, 'name')
    source = chute_find_field(chute, 'source')
    config = chute.get('config', {})

    chute_resolve_source(source, config)

    client = ControllerClient()
    result = client.find_chute(name)
    if result is None:
        raise Exception("Could not find ID for chute {} - is it registered?".format(name))

    result = client.create_version(name, config)
    click.echo(util.format_result(result))
    return result


@root.command('describe-chute')
@click.argument('name')
@click.pass_context
def describe_chute(ctx, name):
    """
    Show detailed information about a chute in the store.

    NAME must be the name of a chute in the store.
    """
    client = ControllerClient()
    result = client.find_chute(name)
    click.echo(util.format_result(result))
    return result


@root.command('help')
@click.pass_context
def help(ctx):
    """
    Show this message and exit.
    """
    click.echo(ctx.parent.get_help())


@root.command('install-chute')
@click.argument('chute')
@click.argument('node')
@click.option('--version', '-v',
        help='Version of the chute to install.')
@click.pass_context
def install_chute(ctx, chute, node, version):
    """
    Install a chute from the store.

    CHUTE must be the name of a chute in the store. NODE must be the
    name of a node that you control.
    """
    client = ControllerClient()
    result = client.install_chute(chute, node, select_version=version)
    click.echo(util.format_result(result))
    return result


@root.command('list-chutes')
@click.pass_context
def list_chutes(ctx):
    """
    List chutes in the store that you own or have access to.
    """
    client = ControllerClient()
    result = client.list_chutes()
    if len(result) > 0:
        click.echo("Name                             Ver Description")
    for chute in sorted(result, key=operator.itemgetter('name')):
        click.echo("{name:32s} {current_version:3d} {description}".format(**chute))
    return result


@root.command('list-versions')
@click.argument('name')
@click.pass_context
def list_versions(ctx, name):
    """
    List versions of a chute in the store.

    NAME must be the name of a chute in the store.
    """
    client = ControllerClient()
    result = client.list_versions(name)
    if len(result) > 0:
        click.echo("Version GitCheckout")
    for version in sorted(result, key=operator.itemgetter('version')):
        try:
            code = version['config']['download']['checkout']
        except:
            code = "N/A"
        click.echo("{:7s} {}".format(str(version['version']), code))
    return result


@root.command()
@click.pass_context
@click.option('--public/--not-public', default=False,
        help='List the chute publicly for other users to download.')
def register(ctx, public):
    """
    Register a chute with the store.

    The chute information including name will be taken from the
    paradrop.yaml file in the current working directory.  If you
    receive an error, it may be that a chute with the same name is
    already registered.
    """
    if not os.path.exists("paradrop.yaml"):
        raise Exception("No paradrop.yaml file found in working directory.")

    with open('paradrop.yaml', 'r') as source:
        chute = yaml.safe_load(source)

    name = chute_find_field(chute, 'name')
    description = chute_find_field(chute, 'description')

    click.echo("Name: {}".format(name))
    click.echo("Description: {}".format(description))
    click.echo("Public: {}".format(public))
    click.echo("")

    prompt = "Ready to send this information to {} (Y/N)? ".format(
            ctx.obj['pdserver_url'])
    if click.confirm(prompt):
        client = ControllerClient()
        result = client.create_chute(name, description, public=public)
        click.echo(util.format_result(result))
    else:
        click.echo("Operation cancelled.")



import getpass
import os
import tempfile

import builtins
import yaml


def format_result(data):
    """
    Format a result from an API call for printing.
    """
    if data is None or data == []:
        return ""
    return yaml.safe_dump(data, default_flow_style=False)


def open_text_editor(data):
    if data is None:
        data = ""

    fd, path = tempfile.mkstemp()
    os.close(fd)

    with open(path, 'w') as output:
        output.write(data)

    # Get modified time before calling editor.
    orig_mtime = os.path.getmtime(path)

    editor = os.environ.get("EDITOR", "vim")
    os.spawnvpe(os.P_WAIT, editor, [editor, path], os.environ)

    with open(path, 'r') as source:
        data = source.read()

    # Check if the file has been modified, and if it has, send the update.
    new_mtime = os.path.getmtime(path)
    if new_mtime == orig_mtime:
        data = None

    os.remove(path)
    return data


def open_yaml_editor(data, description):
    if data is None:
        data = {}

    fd, path = tempfile.mkstemp()
    os.close(fd)

    with open(path, 'w') as output:
        if len(data) > 0:
            output.write(yaml.safe_dump(data, default_flow_style=False))
        output.write("\n")
        output.write("# You are editing the configuration for the {}.\n".format(description))
        output.write("# Blank lines and lines starting with '#' will be ignored.\n")
        output.write("# Save and exit to apply changes; exit without saving to discard.\n")

    # Get modified time before calling editor.
    orig_mtime = os.path.getmtime(path)

    editor = os.environ.get("EDITOR", "vim")
    os.spawnvpe(os.P_WAIT, editor, [editor, path], os.environ)

    with open(path, 'r') as source:
        data = source.read()
        new_data = yaml.safe_load(data)

    # If result is null, convert to an empty dict before sending to router.
    if new_data is None:
        new_data = {}

    # Check if the file has been modified.
    new_mtime = os.path.getmtime(path)
    changed = (new_mtime != orig_mtime)

    os.remove(path)
    return new_data, changed


def update_object(obj, path, callback=None):
    """
    Traverse a data structure ensuring all nodes exist.

    obj: expected to be a dictionary
    path: string with dot-separated path components
    callback: optional callback function (described below)

    When update_object reaches the parent of the leaf node, it calls the
    optional callback function. The arguments to the callback function are:

    - parent: dictionary containing the leaf node
    - key: string key for the leaf node in parent
    - created: boolean flag indicating whether any part of the path, including
        the leaf node needed to be created.

    If the callback function is None, update_object will still ensure that all
    components along the path exist. If the leaf needs to be created, it will
    be created as an empty dictionary.

    Example:
    update_object({}, 'foo.bar') -> {'foo': {'bar': {}}}

    Return value: Returns either the return value of callback, or if callback
    is None, returns the value of the leaf node.
    """
    parts = path.split(".")

    current = obj
    parent = obj
    created = False
    for part in parts:
        if len(part) == 0:
            raise Exception("Path ({}) is invalid".format(path))

        if not isinstance(current, dict):
            raise Exception("Cannot set {}, not a dictionary".format(path))

        # Create dictionaries along the way if path nodes do not exist,
        # but make note of the fact that the previous value did not exist.
        if part not in current:
            current[part] = {}
            created = True

        parent = current
        current = parent[part]

    if callback is not None:
        return callback(parent, parts[-1], created)
    else:
        return current


class LoginGatherer(object):
    """
    LoginGatherer is an iterator that produces username/password tuples.

    On the first iteration, it returns a default username/password combination
    for convenience. On subsequent iterations, it prompts the user for input.

    The class method prompt() can be used directly in a loop for situations
    where the default is not desired.

    Usage examples:

    for username, password in LoginGatherer(netloc):
        ...

    while True:
        username, password = LoginGatherer.prompt(netloc)
    """
    def __init__(self, netloc):
        self.first = True
        self.netloc = netloc

    def __iter__(self):
        self.first = True
        return self

    def __next__(self):
        if self.first:
            self.first = False
            return (LOCAL_DEFAULT_USERNAME, LOCAL_DEFAULT_PASSWORD)
        else:
            return LoginGatherer.prompt(self.netloc)

    def next(self):
        """
        Get the next username and password pair.
        """
        return self.__next__()

    @classmethod
    def prompt(cls, netloc):
        """
        Prompt the user to enter a username and password.

        The netloc argument is presented in the prompt, so that the user knows
        the relevant authentication domain.
        """
        print("Please enter your username and password for {}."
              .format(netloc))
        username = builtins.input("Username: ")
        password = getpass.getpass("Password: ")
        return (username, password)



import datetime
import getpass
import json
import operator
import os
import tarfile
import tempfile
from six.moves.urllib.parse import urlparse

import arrow
import builtins
import click
import yaml

from .comm import change_json, router_login, router_logout, router_request, router_ws_request


@click.group()
@click.argument('address')
@click.pass_context
def device(ctx, address):
    """
    (deprecated) Sub-tree for configuring a device.

    These commands are deprecated. Please use the equivalent commands under
    `pdtools node --help`.
    """
    if address.startswith("http"):
        parts = urlparse(address)
        ctx.obj['address'] = parts.netloc

        if parts.scheme != 'http':
            print("Warning: when specifying the Paradrop device address, "
                  "using a scheme ({}) other than http may result in errors."
                  .format(parts.scheme))

        if not parts.path:
            path = "/api/v1"
        else:
            print("Warning: when specifying the Paradrop device address, "
                  "using a path ({}) other than /api/v1 may result in errors."
                  .format(parts.scheme))
            path = parts.path

        ctx.obj['base_url'] = "{}://{}{}".format(parts.scheme, parts.netloc, path)
        print(ctx.obj['base_url'])

    else:
        ctx.obj['address'] = address
        ctx.obj['base_url'] = "http://{}/api/v1".format(address)


@device.group()
@click.pass_context
def audio(ctx):
    """
    Control device audio properties.
    """
    url = "{}/audio".format(ctx.obj['base_url'])
    ctx.obj['audio_url'] = url


@audio.command()
@click.pass_context
def info(ctx):
    """
    Get audio server information.
    """
    url = "{}/info".format(ctx.obj['audio_url'])
    router_request("GET", url)


@audio.command()
@click.pass_context
def modules(ctx):
    """
    List loaded modules.
    """
    url = "{}/modules".format(ctx.obj['audio_url'])
    router_request("GET", url)


@audio.command('load-module')
@click.pass_context
@click.argument('name')
def load_module(ctx, name):
    """
    Load a module.
    """
    url = "{}/modules".format(ctx.obj['audio_url'])
    data = { 'name': name }
    router_request("POST", url, json=data)


@audio.command()
@click.pass_context
def sinks(ctx):
    """
    List audio sinks.
    """
    url = "{}/sinks".format(ctx.obj['audio_url'])
    router_request("GET", url)


@audio.group()
@click.pass_context
@click.argument('sink_name')
def sink(ctx, sink_name):
    """
    Configure audio sink.
    """
    url = "{}/sinks/{}".format(ctx.obj['audio_url'], sink_name)
    ctx.obj['sink_url'] = url


@sink.command()
@click.pass_context
@click.argument('channel_volume', nargs=-1)
def volume(ctx, channel_volume):
    """
    Set sink volume.
    """
    url = "{}/volume".format(ctx.obj['sink_url'])

    # Convert to a list of floats. Be aware: the obvious approach
    # list(channel_volume) behaves strangely.
    data = [float(vol) for vol in channel_volume]

    router_request("PUT", url, json=data)


@audio.command()
@click.pass_context
def sources(ctx):
    """
    List audio sources.
    """
    url = "{}/sources".format(ctx.obj['audio_url'])
    router_request("GET", url)


@audio.group()
@click.pass_context
@click.argument('source_name')
def source(ctx, source_name):
    """
    Configure audio source.
    """
    url = "{}/sources/{}".format(ctx.obj['audio_url'], source_name)
    ctx.obj['source_url'] = url


@source.command()
@click.pass_context
@click.argument('channel_volume', nargs=-1)
def volume(ctx, channel_volume):
    """
    Set source volume.
    """
    url = "{}/volume".format(ctx.obj['source_url'])

    # Convert to a list of floats. Be aware: the obvious approach
    # list(channel_volume) behaves strangely.
    data = [float(vol) for vol in channel_volume]

    router_request("PUT", url, json=data)


@device.command()
@click.pass_context
def changes(ctx):
    """
    List changes in the working queue.
    """
    url = "{}/changes/".format(ctx.obj['base_url'])
    router_request("GET", url)


@device.command()
@click.pass_context
@click.argument('change_id')
def watch(ctx, change_id):
    """
    Stream messages for a change in progress.
    """
    url = "ws://{}/ws/changes/{}/stream".format(ctx.obj['address'], change_id)

    def on_message(ws, message):
        data = json.loads(message)
        time = datetime.datetime.fromtimestamp(data['time'])
        msg = data['message'].rstrip()
        print("{}: {}".format(time, msg))

    router_ws_request(url, on_message=on_message)


@device.command()
@click.pass_context
def changes(ctx):
    """
    List changes in the working queue.
    """
    url = "{}/changes/".format(ctx.obj['base_url'])
    router_request("GET", url)


@device.group()
@click.pass_context
def chutes(ctx):
    """
    View or create chutes on the router.
    """
    pass


@chutes.command()
@click.pass_context
def list(ctx):
    """
    List chutes installed on the router.
    """
    url = "{}/chutes/".format(ctx.obj['base_url'])
    router_request("GET", url)


@chutes.command()
@click.pass_context
def create(ctx):
    """
    Install a chute from the working directory.
    """
    url = "{}/chutes/".format(ctx.obj['base_url'])
    headers = {'Content-Type': 'application/x-tar'}

    if not os.path.exists("paradrop.yaml"):
        raise Exception("No paradrop.yaml file found in working directory.")

    with tempfile.TemporaryFile() as temp:
        tar = tarfile.open(fileobj=temp, mode="w")
        for dirName, subdirList, fileList in os.walk('.'):
            for fname in fileList:
                path = os.path.join(dirName, fname)
                arcname = os.path.normpath(path)
                tar.add(path, arcname=arcname)
        tar.close()

        temp.seek(0)
        res = router_request("POST", url, headers=headers, data=temp)
        data = res.json()
        ctx.invoke(watch, change_id=data['change_id'])


@device.group()
@click.argument('chute')
@click.pass_context
def chute(ctx, chute):
    """
    Sub-tree for configuring a chute.
    """
    ctx.obj['chute'] = chute
    ctx.obj['chute_url'] = "{}/chutes/{}".format(ctx.obj['base_url'], chute)


@chute.command()
@click.pass_context
def start(ctx):
    """
    Start the chute.
    """
    url = ctx.obj['chute_url'] + "/start"
    res = router_request("POST", url)
    data = res.json()
    ctx.invoke(watch, change_id=data['change_id'])


@chute.command()
@click.pass_context
def stop(ctx):
    """
    Stop the chute.
    """
    url = ctx.obj['chute_url'] + "/stop"
    res = router_request("POST", url)
    data = res.json()
    ctx.invoke(watch, change_id=data['change_id'])


@chute.command()
@click.pass_context
def restart(ctx):
    """
    Restart the chute.
    """
    url = ctx.obj['chute_url'] + "/restart"
    res = router_request("POST", url)
    data = res.json()
    ctx.invoke(watch, change_id=data['change_id'])


@chute.command('edit-environment')
@click.pass_context
def edit_environment(ctx):
    """
    Interactively edit the chute environment vairables.
    """
    req = router_request("GET", ctx.obj['chute_url'], dump=False)
    info = req.json()
    old_environ = info.get('environment', None)
    if old_environ is None:
        old_environ = {}

    fd, path = tempfile.mkstemp()
    os.close(fd)

    with open(path, 'w') as output:
        if len(old_environ) > 0:
            output.write(yaml.safe_dump(old_environ, default_flow_style=False))
        output.write("\n")
        output.write("# You are editing the environment variables for the chute {}.\n"
                .format(ctx.obj['chute']))
        output.write("# Blank lines and lines starting with '#' will be ignored.\n")
        output.write("# Put each variable on a line with a colon separator, e.g. 'VARIABLE: VALUE'\n")
        output.write("# Save and exit to apply changes; exit without saving to discard.\n")

    # Get modified time before calling editor.
    orig_mtime = os.path.getmtime(path)

    editor = os.environ.get("EDITOR", "vim")
    os.spawnvpe(os.P_WAIT, editor, [editor, path], os.environ)

    with open(path, 'r') as source:
        data = source.read()
        new_environ = yaml.safe_load(data)

    # If result is null, convert to an empty dict before sending to router.
    if new_environ is None:
        new_environ = {}

    # Check if the file has been modified, and if it has, send the update.
    new_mtime = os.path.getmtime(path)
    if new_mtime != orig_mtime:
        data = {
            'environment': new_environ
        }
        url = ctx.obj['chute_url'] + "/restart"
        res = router_request("POST", url, json=data)
        data = res.json()
        ctx.invoke(watch, change_id=data['change_id'])

    os.remove(path)


@chute.command()
@click.pass_context
def delete(ctx):
    """
    Uninstall the chute.
    """
    url = ctx.obj['chute_url']
    res = router_request("DELETE", url)
    data = res.json()
    ctx.invoke(watch, change_id=data['change_id'])


@chute.command()
@click.pass_context
def info(ctx):
    """
    Get information about the chute.
    """
    url = ctx.obj['chute_url']
    router_request("GET", url)


@chute.command()
@click.pass_context
def cache(ctx):
    """
    Get details from the chute installation.
    """
    url = ctx.obj['chute_url'] + "/cache"
    router_request("GET", url)


@chute.command()
@click.pass_context
def logs(ctx):
    """
    Watch log messages from a chute.
    """
    url = "ws://{}/sockjs/logs/{}/websocket".format(ctx.obj['address'], ctx.obj['chute'])

    def on_message(ws, message):
        data = json.loads(message)
        time = arrow.get(data['timestamp']).to('local').datetime
        msg = data['message'].rstrip()
        print("{}: {}".format(time, msg))

    router_ws_request(url, on_message=on_message)


@chute.command()
@click.pass_context
def update(ctx):
    """
    Update the chute from the working directory.
    """
    url = ctx.obj['chute_url']
    headers = {'Content-Type': 'application/x-tar'}

    if not os.path.exists("paradrop.yaml"):
        raise Exception("No paradrop.yaml file found in working directory.")

    with tempfile.TemporaryFile() as temp:
        tar = tarfile.open(fileobj=temp, mode="w")
        for dirName, subdirList, fileList in os.walk('.'):
            for fname in fileList:
                path = os.path.join(dirName, fname)
                arcname = os.path.normpath(path)
                tar.add(path, arcname=arcname)
        tar.close()

        temp.seek(0)
        res = router_request("PUT", url, headers=headers, data=temp)
        data = res.json()
        ctx.invoke(watch, change_id=data['change_id'])


@chute.command()
@click.pass_context
def config(ctx):
    """
    Get the chute's current configuration.
    """
    url = ctx.obj['chute_url'] + "/config"
    headers = {'Content-Type': 'application/x-tar'}

    res = router_request("GET", url)


@chute.command()
@click.pass_context
def reconfigure(ctx):
    """
    Reconfigure the chute without rebuilding.
    """
    url = ctx.obj['chute_url'] + "/config"
    headers = {'Content-Type': 'application/x-tar'}

    if not os.path.exists("paradrop.yaml"):
        raise Exception("No paradrop.yaml file found in working directory.")

    with open("paradrop.yaml", "r") as source:
        data = yaml.safe_load(source)
        config = data.get('config', {})

    res = router_request("PUT", url, json=config)
    data = res.json()
    ctx.invoke(watch, change_id=data['change_id'])


@chute.command()
@click.pass_context
def networks(ctx):
    """
    List the chute's networks.
    """
    url = ctx.obj['chute_url'] + "/networks"
    router_request("GET", url)


@chute.command()
@click.pass_context
def shell(ctx):
    """
    Open a shell inside a chute.

    This requires you to have enabled SSH access to the device and installed
    bash inside your chute.
    """
    cmd = ["ssh", "-t", "paradrop@{}".format(ctx.obj['address']), "sudo", "docker",
            "exec", "-it", ctx.obj['chute'], "/bin/bash"]
    os.spawnvpe(os.P_WAIT, "ssh", cmd, os.environ)


@chute.group()
@click.pass_context
@click.argument('network')
def network(ctx, network):
    """
    Sub-tree for accessing chute network.
    """
    ctx.obj['network'] = network
    ctx.obj['network_url'] = "{}/networks/{}".format(ctx.obj['chute_url'],
            network)


@network.command()
@click.pass_context
def stations(ctx):
    """
    List stations connected to the network.
    """
    url = ctx.obj['network_url'] + "/stations"
    router_request("GET", url)


@network.group()
@click.pass_context
@click.argument('station')
def station(ctx, station):
    """
    Sub-tree for accessing network stations.
    """
    ctx.obj['station'] = station
    ctx.obj['station_url'] = ctx.obj['network_url'] + "/stations/" + station


@station.command()
@click.pass_context
def show(ctx):
    """
    Show station information.
    """
    router_request("GET", ctx.obj['station_url'])


@station.command()
@click.pass_context
def delete(ctx):
    """
    Kick a station off the network.
    """
    router_request("DELETE", ctx.obj['station_url'])


@device.group(invoke_without_command=True)
@click.pass_context
def hostconfig(ctx):
    """
    Sub-tree for the host configuration.
    """
    url = ctx.obj['base_url'] + "/config/hostconfig"
    router_request("GET", url)


@device.command()
@click.pass_context
def login(ctx):
    """
    Log in using device-local credentials.
    """
    username = router_login(ctx.obj['base_url'])
    if username is not None:
        print("Logged in as: {}".format(username))
    else:
        print("Log in failed.")


@device.command()
@click.pass_context
def logout(ctx):
    """
    Log out by removing any stored access tokens.
    """
    removed = router_logout(ctx.obj['base_url'])
    print("Removed {} token(s).".format(removed))


@hostconfig.command()
@click.pass_context
@click.argument('option')
@click.argument('value')
def change(ctx, option, value):
    """
    Change one setting in the host configuration.
    """
    url = ctx.obj['base_url'] + "/config/hostconfig"
    req = router_request("GET", url, dump=False)
    config = req.json()

    change_json(config, option, value)
    data = {
        'config': config
    }
    router_request("PUT", url, json=data)


@hostconfig.command()
@click.pass_context
def edit(ctx):
    """
    Interactively edit the host configuration.
    """
    url = ctx.obj['base_url'] + "/config/hostconfig"
    req = router_request("GET", url, dump=False)
    config = req.json()

    fd, path = tempfile.mkstemp()
    os.close(fd)

    with open(path, 'w') as output:
        output.write(yaml.safe_dump(config, default_flow_style=False))

    # Get modified time before calling editor.
    orig_mtime = os.path.getmtime(path)

    editor = os.environ.get("EDITOR", "vim")
    os.spawnvpe(os.P_WAIT, editor, [editor, path], os.environ)

    with open(path, 'r') as source:
        data = source.read()
        config = yaml.safe_load(data)

    # Check if the file has been modified, and if it has, send the update.
    new_mtime = os.path.getmtime(path)
    if new_mtime != orig_mtime:
        data = {
            'config': config
        }
        res = router_request("PUT", url, json=data)
        result = res.json()
        ctx.invoke(watch, change_id=result['change_id'])

    os.remove(path)


@device.group()
@click.pass_context
@click.option('--user', default='paradrop')
def sshkeys(ctx, user):
    """
    Sub-tree for accessing SSH authorized keys.
    """
    ctx.obj['sshkeys_user'] = user
    ctx.obj['sshkeys_url'] = ctx.obj['base_url'] + '/config/sshKeys'


@sshkeys.command()
@click.pass_context
@click.argument('path')
def add(ctx, path):
    """
    Add an authorized key from a file.
    """
    url = '{sshkeys_url}/{sshkeys_user}'.format(**ctx.obj)
    with open(path, 'r') as source:
        key_string = source.read().strip()
        data = {
            'key': key_string
        }

        result = router_request("POST", url, json=data, dump=False)
        if result.ok:
            data = result.json()
            print("Added: " + data.get('key', ''))


@sshkeys.command()
@click.pass_context
def list(ctx):
    """
    List authorized keys.
    """
    url = '{sshkeys_url}/{sshkeys_user}'.format(**ctx.obj)
    result = router_request("GET", url, dump=False)
    if result.ok:
        keys = result.json()
        for key in keys:
            print(key)


@device.command()
@click.pass_context
def password(ctx):
    """
    Change the router admin password.
    """
    url = ctx.obj['base_url'] + "/password/change"

    username = builtins.input("Username: ")
    while True:
        password = getpass.getpass("New password: ")
        confirm = getpass.getpass("Confirm password: ")

        if password == confirm:
            break
        else:
            print("Passwords do not match.")

    data = {
        "username": username,
        "password": password
    }
    router_request("POST", url, json=data)


@device.command()
@click.pass_context
@click.argument('router_id')
@click.argument('router_password')
@click.option('--server', default="https://paradrop.org")
@click.option('--wamp', default="ws://paradrop.org:9086/ws")
def provision(ctx, router_id, router_password, server, wamp):
    """
    Provision the router.
    """
    url = ctx.obj['base_url'] + "/config/provision"

    data = {
        'routerId': router_id,
        'apitoken': router_password,
        'pdserver': server,
        'wampRouter': wamp
    }
    router_request("POST", url, json=data)


def print_pdconf(res):
    if res.ok:
        data = res.json()
        data.sort(key=operator.itemgetter("age"))

        print("{:3s} {:12s} {:20s} {:30s} {:5s}".format("Age", "Type", "Name",
            "Comment", "Pass"))
        for item in data:
            print("{age:<3d} {type:12s} {name:20s} {comment:30s} {success}".format(**item))


@device.group()
@click.pass_context
def pdconf(ctx):
    """
    Access the pdconf subsystem.

    pdconf manages low-level configuration of the Paradrop device.
    These commands are implemented for debugging purposes and are
    not intended for ordinary configuration purposes.
    """
    pass


@pdconf.command()
@click.pass_context
def show(ctx):
    """
    Show status of pdconf subsystem.
    """
    url = ctx.obj['base_url'] + "/config/pdconf"
    res = router_request("GET", url, dump=False)
    print_pdconf(res)


@pdconf.command()
@click.pass_context
def reload(ctx):
    """
    Force pdconf to reload files.
    """
    url = ctx.obj['base_url'] + "/config/pdconf"
    res = router_request("PUT", url, dump=False)
    print_pdconf(res)


@device.group()
@click.pass_context
def snapd(ctx):
    """
    Access the snapd subsystem.
    """
    ctx.obj['snapd_url'] = "http://{}/snapd".format(ctx.obj['address'])


@snapd.command()
@click.pass_context
@click.argument('email')
def createUser(ctx, email):
    """
    Create user account.
    """
    url = ctx.obj['snapd_url'] + "/v2/create-user"
    data = {
        'email': email,
        'sudoer': True
    }
    router_request("POST", url, json=data)


@snapd.command()
@click.pass_context
def connectAll(ctx):
    """
    Connect all interfaces.
    """
    url = ctx.obj['snapd_url'] + "/v2/interfaces"
    res = router_request("GET", url, dump=False)
    data = res.json()
    for item in data['result']['plugs']:
        connections = item.get('connections', [])
        if len(connections) > 0:
            continue

        if item['plug'] == 'docker':
            # The docker slot needs to be treated differently from core slots.
            slot = {'snap': 'docker'}
        elif item['plug'] == 'zerotier-control':
            slot = {'snap': 'zerotier-one'}
        else:
            # Most slots are provided by the core snap and specified this way.
            slot = {'slot': item['interface']}

        req = {
            'action': 'connect',
            'slots': [slot],
            'plugs': [{'snap': item['snap'], 'plug': item['plug']}]
        }
        print("snap connect {snap}:{plug} {interface}".format(**item))
        router_request("POST", url, json=req, dump=False)



"""
This module defines click commands for managing a Paradrop edge compute
node.

One important point about how these commands are implemented is that
the target node is an optional argument. If the target is not specified,
it defaults to the local machine, which means these commands can be run
conveniently on a Paradrop node. It is also possible to override the
default target by setting the PDTOOLS_NODE_TARGET environment variable.
"""
import datetime
import getpass
import json
import operator
import os
import re
import tarfile
import tempfile
from six.moves.urllib.parse import urlparse

import arrow
import builtins
import click
import yaml

from . import util
from .comm import router_ws_request
from .paradrop_client import ParadropClient


def get_base_url(target):
    if target.startswith("http"):
        parts = urlparse(target)

        if parts.scheme != 'http':
            print("Warning: when specifying the Paradrop device address, "
                  "using a scheme ({}) other than http may result in errors."
                  .format(parts.scheme))

        if not parts.path:
            path = "/api/v1"
        else:
            print("Warning: when specifying the Paradrop device address, "
                  "using a path ({}) other than /api/v1 may result in errors."
                  .format(parts.scheme))
            path = parts.path

        return "{}://{}{}".format(parts.scheme, parts.netloc, path)

    else:
        return "http://{}/api/v1".format(target)


def print_pdconf(data):
    data.sort(key=operator.itemgetter("age"))

    print("{:3s} {:12s} {:20s} {:30s} {:5s}".format("Age", "Type", "Name",
        "Comment", "Pass"))
    for item in data:
        print("{age:<3d} {type:12s} {name:20s} {comment:30s} {success}".format(**item))


@click.group('node')
@click.option('--target', '-t', default="172.17.0.1",
        help='Target node name or address')
@click.pass_context
def root(ctx, target):
    """
    Manage a Paradrop edge compute node.

    These commands respect the following environment variables:

    PDTOOLS_NODE_TARGET            Default target node name or address.
    """
    ctx.obj['target'] = target


@root.command('connect-snap-interfaces')
@click.pass_context
def connect_snap_interfaces(ctx):
    """
    Connect all interfaces for installed snaps.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.list_snap_interfaces()
    for item in result['result']['plugs']:
        connections = item.get('connections', [])
        if len(connections) > 0:
            continue

        if item['plug'] == 'docker':
            # The docker slot needs to be treated differently from core slots.
            slot = {'snap': 'docker'}
        elif item['plug'] == 'zerotier-control':
            # TODO: This connection is failing, but I am not sure why.
            slot = {'snap': 'zerotier-one', 'slot': 'zerotier-control'}
        else:
            # Most slots are provided by the core snap and specified this way.
            slot = {'slot': item['interface']}

        result = client.connect_snap_interface(slots=[slot], plugs=[{'snap':
            item['snap'], 'plug': item['plug']}])
        if result['type'] == 'error':
            print(result['result']['message'])


@root.command('create-user')
@click.argument('email')
@click.pass_context
def create_user(ctx, email):
    """
    Create local Linux user connected to Ubuntu store account.

    EMAIL must be an email address which is registered as Ubuntu One
    account.  The name of the local account that is created will depend
    on the email address used. If in doubt, use "info@paradrop.io",
    which will result in a user named "paradrop" being created.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.create_user(email)
    click.echo(util.format_result(result))
    return result


@root.command('describe-audio')
@click.pass_context
def describe_audio(ctx):
    """
    Display audio subsystem information.

    Display information from the local PulseAudio server such as the
    default source and sink.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.get_audio()
    click.echo(util.format_result(result))
    return result


@root.command('describe-chute')
@click.argument('chute')
@click.pass_context
def describe_chute(ctx, chute):
    """
    Display information about a chute.

    CHUTE must be the name of an installed chute.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.get_chute(chute)
    click.echo(util.format_result(result))
    return result


@root.command('describe-chute-cache')
@click.argument('chute')
@click.pass_context
def describe_chute_cache(ctx, chute):
    """
    Show internal details from a chute installation.

    CHUTE must be the name of an installed chute.

    This information is intended for Paradrop daemon developers for
    debugging purposes. The output is not expected to remain stable
    between Paradrop versions.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.get_chute_cache(chute)
    click.echo(util.format_result(result))
    return result


@root.command('describe-chute-configuration')
@click.argument('chute')
@click.pass_context
def describe_chute_configuration(ctx, chute):
    """
    Display configuration of a chute.

    CHUTE must be the name of an installed chute.

    This information corresponds to the "config" section in a chute's
    paradrop.yaml file.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.get_chute_config(chute)
    click.echo(util.format_result(result))
    return result


@root.command('describe-chute-network-client')
@click.argument('chute')
@click.argument('network')
@click.argument('client')
@click.pass_context
def describe_chute_network_client(ctx, chute, network, client):
    """
    Display information about a network client.

    CHUTE must be the name of an installed chute.  NETWORK must be the
    name of one of the chute's configured networks. Typically, this
    will be "wifi".  CLIENT identifies the network client, such as a
    MAC address.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.get_chute_client(chute, network, client)
    click.echo(util.format_result(result))
    return result


@root.command('describe-pdconf')
@click.pass_context
def describe_pdconf(ctx):
    """
    Show status of the pdconf subsystem.

    This information is intended for Paradrop daemon developers for
    debugging purposes.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.get_pdconf()
    print_pdconf(result)
    return result


@root.command('describe-provision')
@click.pass_context
def describe_provision(ctx):
    """
    Show provisioning status of the node.

    This shows whether the node is associated with a cloud controller,
    and if so, its identifier.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.get_provision()
    click.echo(util.format_result(result))
    return result


@root.command('edit-configuration')
@click.pass_context
def edit_configuration(ctx):
    """
    Interactively edit the node configuration and apply changes.

    Open the text editor specified by the EDITOR environment variable
    with the current node configuration. If you save and exit, the new
    configuration will be applied to the node.
    """
    client = ParadropClient(ctx.obj['target'])
    old_data = client.get_config()
    new_data, changed = util.open_yaml_editor(old_data, "node")
    if changed:
        result = client.set_config(new_data)
        ctx.invoke(watch_change_logs, change_id=result['change_id'])
    return new_data


@root.command('edit-chute-configuration')
@click.argument('chute')
@click.pass_context
def edit_chute_configuration(ctx, chute):
    """
    Interactively edit the chute configuration and restart it.

    CHUTE must be the name of an installed chute.

    Open the text editor specified by the EDITOR environment variable
    with the current chute configuration. If you save and exit, the new
    configuration will be applied and the chute restarted.
    """
    client = ParadropClient(ctx.obj['target'])
    old_data = client.get_chute_config(chute)
    new_data, changed = util.open_yaml_editor(old_data, "chute " + chute)
    if changed:
        result = client.set_chute_config(chute, new_data)
        ctx.invoke(watch_change_logs, change_id=result['change_id'])
    return new_data


@root.command('edit-chute-variables')
@click.argument('chute')
@click.pass_context
def edit_chute_variables(ctx, chute):
    """
    Interactively edit a chute's environment variables and restart it.

    CHUTE must be the name of an installed chute.

    Open the text editor specified by the EDITOR environment variable
    with the current chute environment variables. If you save and exit,
    the new settings will be applied and the chute restarted.
    """
    client = ParadropClient(ctx.obj['target'])
    old_data = client.get_chute(chute).get('environment', {})
    new_data, changed = util.open_yaml_editor(old_data, "chute " + chute)
    if changed:
        result = client.set_chute_variables(chute, new_data)
        ctx.invoke(watch_change_logs, change_id=result['change_id'])
    return new_data


@root.command('export-configuration')
@click.option('--format', '-f', default='json',
        help="Format (json or yaml)")
@click.pass_context
def export_configuration(ctx, format):
    """
    Display the node configuration in the desired format.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.get_config()

    format = format.lower()
    if format == 'json':
        click.echo(json.dumps(result, indent=4))
    elif format == 'yaml':
        click.echo(yaml.safe_dump(result, default_flow_style=False))
    else:
        click.echo("Unrecognized format: {}".format(format))
    return result


@root.command('generate-configuration')
@click.option('--format', '-f', default='json',
        help="Format (json or yaml)")
@click.pass_context
def generate_configuration(ctx, format):
    """
    Generate a new node configuration based on detected hardware.

    The new configuration is not automatically applied.  Rather, you can
    save it to file and use the import-configuration command to apply it.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.generate_config()

    format = format.lower()
    if format == 'json':
        click.echo(json.dumps(result, indent=4))
    elif format == 'yaml':
        click.echo(yaml.safe_dump(result, default_flow_style=False))
    else:
        click.echo("Unrecognized format: {}".format(format))
    return result


@root.command('help')
@click.pass_context
def help(ctx):
    """
    Show this message and exit.
    """
    click.echo(ctx.parent.get_help())


@root.command('import-configuration')
@click.argument('path')
@click.pass_context
def import_configuration(ctx, path):
    """
    Import a node configuration from file and apply changes.

    PATH must be a path to a node configuration file in YAML format.
    """
    with open(path, 'r') as source:
        config = yaml.safe_load(source)

    client = ParadropClient(ctx.obj['target'])
    result = client.set_config(config)
    ctx.invoke(watch_change_logs, change_id=result['change_id'])
    return result


@root.command('import-ssh-key')
@click.argument('path')
@click.option('--user', '-u', default='paradrop', help='Local username')
@click.pass_context
def import_ssh_key(ctx, path, user):
    """
    Add an authorized key from a public key file.

    PATH must be a path to a public key file, which corresponds to
    a private key that SSH can use for authentication. Typically,
    ssh-keygen will place the public key in "~/.ssh/id_rsa.pub".
    """
    client = ParadropClient(ctx.obj['target'])
    with open(path, 'r') as source:
        key_string = source.read().strip()

    match = re.search("-----BEGIN \w+ PRIVATE KEY-----", key_string)
    if match is not None:
        print("The path ({}) contains a private key.".format(path))
        print("Please provide the path to your public key.")
        return None

    result = client.add_ssh_key(key_string, user=user)
    if result is not None:
        print("Added public key from {}".format(path))

    return result

@root.command('install-chute')
@click.option('--directory', '-d', default='.', help='Directory containing chute files')
@click.pass_context
def install_chute(ctx, directory):
    """
    Install a chute from the working directory.

    Install the files in the current directory as a chute on the node.
    The directory must contain a paradrop.yaml file.  The entire directory
    will be copied to the node for installation.
    """
    os.chdir(directory)

    if not os.path.exists("paradrop.yaml"):
        raise Exception("No paradrop.yaml file found in chute directory.")

    client = ParadropClient(ctx.obj['target'])
    with tempfile.TemporaryFile() as temp:
        tar = tarfile.open(fileobj=temp, mode="w")
        for dirName, subdirList, fileList in os.walk("."):
            for fname in fileList:
                path = os.path.join(dirName, fname)
                arcname = os.path.normpath(path)
                tar.add(path, arcname=arcname)
        tar.close()

        temp.seek(0)
        result = client.install_tar(temp)
        ctx.invoke(watch_change_logs, change_id=result['change_id'])


@root.command('list-audio-modules')
@click.pass_context
def list_audio_modules(ctx):
    """
    List modules loaded by the audio subsystem.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.list_audio_modules()
    click.echo(util.format_result(result))
    return result


@root.command('list-audio-sinks')
@click.pass_context
def list_audio_sinks(ctx):
    """
    List audio sinks.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.list_audio_sinks()
    click.echo(util.format_result(result))
    return result


@root.command('list-audio-sources')
@click.pass_context
def list_audio_sources(ctx):
    """
    List audio sources.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.list_audio_sources()
    click.echo(util.format_result(result))
    return result


@root.command('list-changes')
@click.pass_context
def list_changes(ctx):
    """
    List queued or in progress changes.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.list_changes()
    click.echo(util.format_result(result))
    return result


@root.command('list-chute-networks')
@click.argument('chute')
@click.pass_context
def list_chute_networks(ctx, chute):
    """
    List networks configured by a chute.

    CHUTE must be the name of an installed chute.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.list_chute_networks(chute)
    click.echo(util.format_result(result))
    return result


@root.command('list-chute-network-clients')
@click.argument('chute')
@click.argument('network')
@click.pass_context
def list_chute_network_clients(ctx, chute, network):
    """
    List clients connected to the chute's network.

    CHUTE must be the name of an installed chute. NETWORK must be the
    name of one of its configured networks. Typically, this will be called
    "wifi".
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.list_chute_clients(chute, network)
    click.echo(util.format_result(result))
    return result


@root.command('list-chutes')
@click.pass_context
def list_chutes(ctx):
    """
    List chutes installed on the node
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.list_chutes()
    click.echo(util.format_result(result))
    return result


@root.command('list-snap-interfaces')
@click.pass_context
def list_snap_interfaces(ctx):
    """
    List interfaces for snaps installed on the node.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.list_snap_interfaces()
    click.echo(util.format_result(result))
    return result


@root.command('list-ssh-keys')
@click.option('--user', '-u', default='paradrop', help='Local username')
@click.pass_context
def list_ssh_keys(ctx, user):
    """
    List keys authorized for SSH access to the node.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.list_ssh_keys(user=user)
    click.echo(util.format_result(result))
    return result


@root.command('load-audio-module')
@click.argument('module')
@click.pass_context
def load_audio_module(ctx, module):
    """
    Load a module into the audio subsystem.

    MODULE must be the name of a PulseAudio module such as
    "module-loopback".
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.load_audio_module(module)
    click.echo(util.format_result(result))
    return result


@root.command('login')
@click.pass_context
def login(ctx):
    """
    Interactively log in using the local admin password.

    Authenticate with the node using the local username and
    password. Typically, the username will be "paradrop". The password
    can be set with the set-password command.
    """
    client = ParadropClient(ctx.obj['target'])
    token = client.login()
    if token is None:
        click.echo("Login attempt failed.")
        return False
    else:
        click.echo("Login successful.")
        return True


@root.command('logout')
@click.pass_context
def logout(ctx):
    """
    Log out and remove stored credentials.
    """
    client = ParadropClient(ctx.obj['target'])
    removed = client.logout()
    click.echo("Removed {} token(s).".format(removed))
    return removed


@root.command('open-chute-shell')
@click.argument('chute')
@click.pass_context
def open_chute_shell(ctx, chute):
    """
    Open a shell inside the running chute.

    CHUTE must be the name of a running chute.

    This requires you to have enabled SSH access to the device and
    installed bash inside your chute.

    Changes made to files inside the chute may not be persistent if the
    chute or the node is restarted. Only changes to files in the "/data"
    directory will be preserved.
    """
    cmd = ["ssh", "-t", "paradrop@{}".format(ctx.obj['target']), "sudo", "docker",
            "exec", "-it", chute, "/bin/bash"]
    os.spawnvpe(os.P_WAIT, "ssh", cmd, os.environ)


@root.command('provision')
@click.argument('id')
@click.argument('key')
@click.option('--controller', '-c', default="https://paradrop.org", help="Cloud controller endpoint")
@click.option('--wamp', '-w', default="ws://paradrop.org:9086/ws", help="WAMP endpoint")
@click.pass_context
def provision(ctx, id, key, controller, wamp):
    """
    Associate the node with a cloud controller.

    ID and KEY are credentials that can be found when creating a node
    on the controller, either through the website or through `pdtools
    cloud create-node`.  They may also be referred to as the Router ID
    and the Router Password.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.provision(id, key, controller=controller, wamp=wamp)
    click.echo(util.format_result(result))
    return result


@root.command('reboot')
@click.pass_context
def reboot(ctx):
    """
    Reboot the node.
    """
    change = {
        'name': '__PARADROP__',
        'updateClass': 'ROUTER',
        'updateType': 'reboot'
    }
    client = ParadropClient(ctx.obj['target'])
    result = client.add_change(change)
    ctx.invoke(watch_change_logs, change_id=result['change_id'])


@root.command('remove-chute')
@click.argument('chute')
@click.pass_context
def remove_chute(ctx, chute):
    """
    Remove a chute from the node.

    CHUTE must be the name of an installed chute.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.remove_chute(chute)
    ctx.invoke(watch_change_logs, change_id=result['change_id'])


@root.command('remove-chute-network-client')
@click.argument('chute')
@click.argument('network')
@click.argument('client')
@click.pass_context
def remove_chute_network_client(ctx, chute, network, client):
    """
    Remove a connected client from the chute's network.

    CHUTE must be the name of an installed chute.  NETWORK must be the
    name of one of the chute's configured networks. Typically, this
    will be "wifi".  CLIENT identifies the network client, such as a
    MAC address.

    Only implemented for wireless clients, this effectively kicks the
    client off the network.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.remove_chute_client(chute, network, client)
    click.echo(util.format_result(result))
    return result


@root.command('restart-chute')
@click.argument('chute')
@click.pass_context
def restart_chute(ctx, chute):
    """
    Restart a chute.

    CHUTE must be the name of an installed chute.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.restart_chute(chute)
    ctx.invoke(watch_change_logs, change_id=result['change_id'])


@root.command('set-configuration')
@click.argument('path')
@click.argument('value')
@click.pass_context
def set_configuration(ctx, path, value):
    """
    Change a node configuration value and apply.

    PATH must be a dot-separated path to a value in the node
    configuration, such as "system.onMissingWiFi". VALUE will be
    interpreted as a string, numeric, or boolean type as appropriate.

    Changing values inside a list is not currently supported.
    Use edit-configuration instead.
    """
    client = ParadropClient(ctx.obj['target'])
    config = client.get_config()

    # This does some type inference for boolean, numerics, etc.
    value = yaml.safe_load(value)

    def set_value(parent, key, created):
        if created:
            print("Created new field {} = {}".format(path, value))
        else:
            current = parent[key]
            print("Changed {} from {} to {}".format(path, current, value))
        parent[key] = value

    util.update_object(config, path, set_value)

    result = client.set_config(config)
    ctx.invoke(watch_change_logs, change_id=result['change_id'])


@root.command('set-sink-volume')
@click.argument('sink')
@click.argument('volume', nargs=-1)
@click.pass_context
def set_sink_volume(ctx, sink, volume):
    """
    Configure audio sink volume.

    SINK must be the name of a PulseAudio sink. VOLUME should be one
    (applied to all channels) or multiple (one for each channel) floating
    point values between 0 and 1.
    """
    client = ParadropClient(ctx.obj['target'])

    # Convert to a list of floats. Be aware: the obvious approach
    # list(volume) behaves strangely.
    data = [float(vol) for vol in volume]

    result = client.set_sink_volume(sink, data)
    click.echo(util.format_result(result))
    return result


@root.command('set-source-volume')
@click.argument('source')
@click.argument('volume', nargs=-1)
@click.pass_context
def set_source_volume(ctx, source, volume):
    """
    Configure audio source volume.

    SOURCE must be the name of a PulseAudio source. VOLUME should be one
    (applied to all channels) or multiple (one for each channel) floating
    point values between 0 and 1.
    """
    client = ParadropClient(ctx.obj['target'])

    # Convert to a list of floats. Be aware: the obvious approach
    # list(volume) behaves strangely.
    data = [float(vol) for vol in volume]

    result = client.set_source_volume(source, data)
    click.echo(util.format_result(result))
    return result


@root.command('set-password')
@click.pass_context
def set_password(ctx):
    """
    Change the local admin password.

    Set the password required by `pdtools node login` and the local
    web-based administration page.
    """
    username = builtins.input("Username: ")
    while True:
        password = getpass.getpass("New password: ")
        confirm = getpass.getpass("Confirm password: ")

        if password == confirm:
            break
        else:
            print("Passwords do not match.")

    click.echo("Next, if prompted, you should enter the current username and password.")
    client = ParadropClient(ctx.obj['target'])
    result = client.set_password(username, password)
    click.echo(util.format_result(result))
    return result


@root.command('shutdown')
@click.pass_context
def shutdown(ctx):
    """
    Shut down the node.
    """
    change = {
        'name': '__PARADROP__',
        'updateClass': 'ROUTER',
        'updateType': 'shutdown'
    }
    client = ParadropClient(ctx.obj['target'])
    result = client.add_change(change)
    ctx.invoke(watch_change_logs, change_id=result['change_id'])


@root.command('start-chute')
@click.argument('chute')
@click.pass_context
def start_chute(ctx, chute):
    """
    Start a stopped chute.

    CHUTE must be the name of a stopped chute.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.start_chute(chute)
    ctx.invoke(watch_change_logs, change_id=result['change_id'])


@root.command('stop-chute')
@click.argument('chute')
@click.pass_context
def stop_chute(ctx, chute):
    """
    Stop a running chute.

    CHUTE must be the name of a running chute.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.stop_chute(chute)
    ctx.invoke(watch_change_logs, change_id=result['change_id'])


@root.command('trigger-pdconf')
@click.pass_context
def trigger_pdconf(ctx):
    """
    Trigger pdconf to reload configuration.

    This function is intended for Paradrop daemon developers for
    debugging purposes. Generally, you should use edit-configuration
    and edit-chute-configuration for making configuration changes.
    """
    client = ParadropClient(ctx.obj['target'])
    result = client.trigger_pdconf()
    print_pdconf(result)


@root.command('update-chute')
@click.option('--directory', '-d', default='.', help='Directory containing chute files')
@click.pass_context
def update_chute(ctx, directory):
    """
    Install a new version of the chute from the working directory.

    Install the files in the current directory as a chute on the node.
    The directory must contain a paradrop.yaml file.  The entire directory
    will be copied to the node for installation.
    """
    os.chdir(directory)

    if not os.path.exists("paradrop.yaml"):
        raise Exception("No paradrop.yaml file found in chute directory.")

    with open('paradrop.yaml', 'r') as source:
        config = yaml.safe_load(source)

    if 'name' not in config:
        click.echo('Chute name is not defined in paradrop.yaml.')
        return

    client = ParadropClient(ctx.obj['target'])
    with tempfile.TemporaryFile() as temp:
        tar = tarfile.open(fileobj=temp, mode="w")
        for dirName, subdirList, fileList in os.walk("."):
            for fname in fileList:
                path = os.path.join(dirName, fname)
                arcname = os.path.normpath(path)
                tar.add(path, arcname=arcname)
        tar.close()

        temp.seek(0)
        result = client.install_tar(temp, name=config['name'])
        ctx.invoke(watch_change_logs, change_id=result['change_id'])


@root.command('watch-change-logs')
@click.argument('change_id')
@click.pass_context
def watch_change_logs(ctx, change_id):
    """
    Stream log messages from an in-progress change.

    CHANGE_ID must be the ID of a queued or in-progress change as
    retrieved from the list-changes command.
    """
    url = "ws://{}/ws/changes/{}/stream".format(ctx.obj['target'], change_id)

    def on_message(ws, message):
        data = json.loads(message)
        time = datetime.datetime.fromtimestamp(data['time'])
        msg = data['message'].rstrip()
        print("{}: {}".format(time, msg))

    router_ws_request(url, on_message=on_message)


@root.command('watch-chute-logs')
@click.argument('chute')
@click.pass_context
def watch_chute_logs(ctx, chute):
    """
    Stream log messages from a running chute.

    CHUTE must be the name of a running chute.
    """
    url = "ws://{}/sockjs/logs/{}/websocket".format(ctx.obj['target'], chute)

    def on_message(ws, message):
        data = json.loads(message)
        time = arrow.get(data['timestamp']).to('local').datetime
        msg = data['message'].rstrip()
        print("{}: {}".format(time, msg))

    router_ws_request(url, on_message=on_message)


@root.command('watch-logs')
@click.pass_context
def watch_logs(ctx):
    """
    Stream log messages from the Paradrop daemon.
    """
    url = "ws://{}/ws/paradrop_logs".format(ctx.obj['target'])

    def on_message(ws, message):
        print(message)

    router_ws_request(url, on_message=on_message)



import os
from six.moves.urllib.parse import urlparse

from .authenticated_client import AuthenticatedClient
from .config import PdtoolsConfig
from .devices.camera import Camera
from .util import LoginGatherer


LOCAL_DEFAULT_USERNAME = "paradrop"
LOCAL_DEFAULT_PASSWORD = ""

PARADROP_API_TOKEN = os.environ.get("PARADROP_API_TOKEN", None)
PARADROP_CHUTE_NAME = os.environ.get("PARADROP_CHUTE_NAME", None)

PDSERVER_URL = os.environ.get('PDSERVER_URL', 'https://paradrop.org')


class ControllerClient(AuthenticatedClient):
    """
    Client for interacting with a cloud controller.
    """
    def __init__(self, host=PDSERVER_URL):
        super(ControllerClient, self).__init__("cloud", PDSERVER_URL)
        self.host = host
        self.base_url = host + "/api"
        #self.base_url = "http://{}/api".format(host)

    def claim_node(self, token):
        """
        Claim ownership of a node using a claim token.
        """
        url = self.base_url + "/routers/claim"
        data = {
            "claim_token": token
        }
        return self.request("POST", url, json=data)

    def create_chute(self, name, description, public=False):
        """
        Create a new chute in the store.
        """
        url = self.base_url + "/chutes"
        data = {
            "name": name,
            "description": description,
            "public": public
        }
        return self.request("POST", url, json=data)

    def create_node(self, name, orphaned=False, claim=None):
        """
        Create a new node tracked by the controller.
        """
        url = self.base_url + "/routers"
        data = {
            "name": name,
            "orphaned": orphaned
        }
        if claim is not None:
            data['claim'] = claim
        return self.request("POST", url, json=data)

    def create_user(self, name, email, password, password2):
        """
        Create a new user account on the controller.
        """
        url = self.base_url + "/users"
        data = {
            "name": name,
            "email": email,
            "password": password,
            "confirmPassword": password2
        }
        return self.request("POST", url, json=data)

    def create_version(self, name, config):
        """
        Create a new chute version.
        """
        chute = self.find_chute(name)
        if chute is None:
            return None

        url = "{}/chutes/{}/versions".format(self.base_url, chute['_id'])
        data = {
            "chute_id": chute['_id'],
            "config": config
        }
        return self.request("POST", url, json=data)

    def delete_node(self, name):
        """
        Delete a node tracked by the controller.
        """
        node = self.find_node(name)
        if node is not None:
            url = "{}/routers/{}".format(self.base_url, node['_id'])
            return self.request("DELETE", url)
        else:
            return None

    def find_chute(self, name):
        """
        Find a chute by name or id.
        """
        # If this client object is ever used for multiple requests during its
        # lifetime, we could consider caching the group list locally for a
        # better response time. Then we need to add cache invalidation to all
        # of the methods that might affect the group list.
        chutes = self.list_chutes()
        for chute in chutes:
            if chute['_id'] == name or chute['name'] == name:
                return chute
        return None

    def find_group(self, name):
        """
        Find a group by name or id.
        """
        # If this client object is ever used for multiple requests during its
        # lifetime, we could consider caching the group list locally for a
        # better response time. Then we need to add cache invalidation to all
        # of the methods that might affect the group list.
        groups = self.list_groups()
        for group in groups:
            if group['_id'] == name or group['name'] == name:
                return group
        return None

    def find_node(self, name):
        """
        Find a node by name or id.
        """
        # If this client object is ever used for multiple requests during its
        # lifetime, we could consider caching the node list locally for a
        # better response time. Then we need to add cache invalidation to all
        # of the methods that might affect the node list.
        nodes = self.list_nodes()
        for node in nodes:
            if node['_id'] == name or node['name'] == name:
                return node
        return None

    def group_add_node(self, group_name, node_name):
        """
        Add a node to a group.
        """
        group = self.find_group(group_name)
        if group is None:
            raise Exception("Group was not found")
        node = self.find_node(node_name)
        if node is None:
            raise Exception("Node was not found")

        url = "{}/groups/{}/addRouter".format(self.base_url, group['_id'])
        data = {
            'router_id': node['_id']
        }
        return self.request("POST", url, json=data)

    def install_chute(self, chute_name, node_name, select_version=None):
        """
        Install a chute from the store.
        """
        chute = self.find_chute(chute_name)
        if chute is None:
            raise Exception("Chute was not found")
        node = self.find_node(node_name)
        if node is None:
            raise Exception("Node was not found")

        versions = self.list_versions(chute_name)
        if versions is None:
            raise Exception("No version to install")

        version = None
        if select_version is None:
            version = versions[-1]
        else:
            for ver in versions:
                if str(ver['version']) == str(select_version):
                    version = ver
                    break
        if version is None:
            raise Exception("Version not found")

        data = {
            "updateClass": "CHUTE",
            "updateType": "update",
            "chute_id": chute['_id'],
            "router_id": node['_id'],
            "version_id": version['_id'],
            "config": version['config']
        }

        # Important: the server will reject the update if the name field is
        # missing. Version is also not automatically filled in.
        data['config']['name'] = chute['name']
        data['config']['version'] = version['version']

        url = "{}/routers/{}/updates".format(self.base_url, node['_id'])
        return self.request("POST", url, json=data)

    def list_chutes(self):
        """
        List chutes that the user owns or has access to.
        """
        url = self.base_url + "/chutes"
        return self.request("GET", url)

    def list_groups(self):
        """
        List groups that the user belongs to.
        """
        url = self.base_url + "/groups"
        return self.request("GET", url)

    def list_nodes(self):
        """
        List nodes that the user owns or has access to.
        """
        url = self.base_url + "/routers"
        return self.request("GET", url)

    def list_versions(self, name):
        """
        List nodes that the user owns or has access to.
        """
        chute = self.find_chute(name)
        if chute is None:
            return []

        url = "{}/chutes/{}/versions".format(self.base_url, chute['_id'])
        return self.request("GET", url)

    def save_node(self, node):
        """
        Save changes to a node object.
        """
        url = "{}/routers/{}".format(self.base_url, node['_id'])
        return self.request("PUT", url, json=node)



import click

from .comm import pdserver_request


@click.command('list-groups')
@click.pass_context
def list_groups(ctx):
    """
    (deprecated) List user groups.

    Please use `pdtools cloud list-groups`.
    """
    url = ctx.obj['pdserver_url'] + "/api/groups"
    result = pdserver_request('GET', url)
    groups = result.json()

    for group in groups:
        print("{} {}".format(group['_id'], group['name']))

        print("  Users:")
        for user_id in group['users']:
            print("    {}".format(user_id))

        print("  Routers:")
        for router_id in group['routers']:
            print("    {}".format(router_id))


@click.group()
@click.pass_context
@click.argument('group_id')
def group(ctx, group_id):
    """
    (deprecated) Manage a user group.

    These commands are deprecated. Please use the equivalent commands under
    `pdtools cloud --help`.
    """
    ctx.obj['group_url'] = ctx.obj['pdserver_url'] + "/api/groups/{}".format(group_id)


@group.command('add-router')
@click.pass_context
@click.argument('router_id')
def add_router(ctx, router_id):
    """
    Add a router to the group.
    """
    url = ctx.obj['group_url'] + "/addRouter"
    data = {
        'router_id': router_id
    }
    result = pdserver_request('POST', url, json=data)
    if result.ok:
        print("Added router.")
    else:
        print("There was an error adding the router.")


def register_commands(root):
    """
    Register commands in this module.
    """
    root.add_command(list_groups)
    root.add_command(group)



import base64

import requests
import six


class Camera(object):
    def __init__(self, device_id, host):
        """
        Initialize Camera device object.

        device_id: string that uniquely identifies the device (e.g. MAC address)
        host: host name or IP address
        """
        self.device_id = device_id
        self.host = host

    def __repr__(self):
        return "Camera({})".format(self.device_id)

    def get_image(self):
        """
        Get an image from the camera.

        Returns image data as a BytesIO/StringIO object.
        """
        url = "http://{}/image.jpg".format(self.host)

        encoded = base64.b64encode('admin:'.encode('utf-8')).decode('ascii')

        headers = {
            'Authorization': 'Basic ' + encoded
        }

        result = requests.get(url, headers=headers)
        if result.ok:
            return six.BytesIO(result.content)

        else:
            return None



import getpass

import builtins
import click

from . import util
from .controller_client import ControllerClient


@click.group('cloud')
@click.pass_context
def root(ctx):
    """
    Access services provided by a cloud controller.

    By default the cloud controller is assumed to be paradrop.org. This
    can be configured through the environment variable PDSERVER_URL.
    """
    pass


@root.command('claim-node')
@click.argument('token')
@click.pass_context
def claim_node(ctx, token):
    """
    Take ownership of a node by using a claim token.

    TOKEN is a hard-to-guess string that the previous owner would have
    configured when setting the node's status as orphaned.
    """
    client = ControllerClient()
    result = client.claim_node(token)
    if result is not None:
        click.echo("Claimed node with name {}".format(result['name']))
    else:
        click.echo("No node was found with that claim token.")
    return result


@root.command('create-node')
@click.argument('name')
@click.option('--orphaned/--not-orphaned', default=False,
        help='Allow another user to claim the node.')
@click.option('--claim', default=None,
        help='Claim token required to claim the node.')
@click.pass_context
def create_node(ctx, name, orphaned, claim):
    """
    Create a new node to be tracked by the controller.

    NAME must be unique among the nodes that you own and may only consist
    of lowercase letters, numbers, and hyphens. It must also begin with
    a letter.
    """
    client = ControllerClient()
    result = client.create_node(name, orphaned=orphaned, claim=claim)
    click.echo(util.format_result(result))
    return result


@root.command('delete-node')
@click.argument('name')
@click.pass_context
def delete_node(ctx, name):
    """
    Delete a node that is tracked by the controller.

    NAME must be the name of a node that you own.
    """
    client = ControllerClient()
    result = client.delete_node(name)
    click.echo(util.format_result(result))
    return result


@root.command('describe-node')
@click.argument('name')
@click.pass_context
def describe_node(ctx, name):
    """
    Get detailed information about an existing node.

    NAME must be the name of a node that you own.
    """
    client = ControllerClient()
    result = client.find_node(name)
    click.echo(util.format_result(result))
    return result


@root.command('edit-node-description')
@click.argument('name')
@click.pass_context
def edit_node_description(ctx, name):
    """
    Interactively edit the node description and save.

    NAME must be the name of a node that you own.

    Open the text editor specified by the EDITOR environment variable
    with the current node description. If you save and exit, the changes
    will be applied to the node.
    """
    client = ControllerClient()
    node = client.find_node(name)
    if node is None:
        click.echo("Node {} was not found.".format(name))
        return

    description = click.edit(node.get('description', ''))
    if description is None:
        click.echo("No change made.")
    else:
        node['description'] = description
        result = client.save_node(node)
        click.echo(util.format_result(result))
        return result


@root.command('group-add-node')
@click.argument('group')
@click.argument('node')
@click.pass_context
def group_add_node(ctx, group, node):
    """
    Add a node to a group for other members to access.

    GROUP must be the string ID of a group. NODE must be the name of a
    node that you control.
    """
    client = ControllerClient()
    result = client.group_add_node(group, node)
    click.echo(util.format_result(result))
    return result


@root.command('help')
@click.pass_context
def help(ctx):
    """
    Show this message and exit.
    """
    click.echo(ctx.parent.get_help())


@root.command('list-groups')
@click.pass_context
def list_groups(ctx):
    """
    List groups that you belong to.
    """
    client = ControllerClient()
    result = client.list_groups()
    click.echo(util.format_result(result))
    return result


@root.command('list-nodes')
@click.pass_context
def list_nodes(ctx):
    """
    List nodes that you own or have access to.
    """
    client = ControllerClient()
    result = client.list_nodes()
    if len(result) > 0:
        click.echo("Name             Online Version Description")
    else:
        click.echo("No nodes were found.")
    for node in result:
        description = node.get('description', '')
        version = node.get('platform_version', 'unknown').split('+')[0]
        click.echo("{name:16s} {online!s:6s} {0:7s} {1}".format(version, description, **node))
    return result


@root.command('login')
@click.pass_context
def login(ctx):
    """
    Interactively login to your user account on the controller.

    Authenticate with the controller using account credentials that you
    created either through the website or with the register command.
    Typically, the username will be your email address.
    """
    client = ControllerClient()
    token = client.login()
    if token is None:
        click.echo("Login attempt failed.")
        return False
    else:
        click.echo("Login successful.")
        return True


@root.command('logout')
@click.pass_context
def logout(ctx):
    """
    Log out and remove stored credentials.
    """
    client = ControllerClient()
    removed = client.logout()
    click.echo("Removed {} token(s).".format(removed))
    return removed


@root.command('register')
@click.pass_context
def register(ctx):
    """
    Interactively create an account on the controller.
    """
    name = builtins.input("Name: ")
    email = builtins.input("Email: ")
    password = getpass.getpass("Password: ")
    password2 = getpass.getpass("Confirm password: ")

    client = ControllerClient()
    result = client.create_user(name, email, password, password2)

    retval = None

    if 'message' in result:
        click.echo(result['message'])

    if 'token' in result:
        click.echo("The account registration was successful. "
                   "Please check your email and verify the account "
                   "before attempting to log in and use any other "
                   "pdtools cloud commands.")
        retval = email

    if 'errors' in result:
        for key, value in result['errors'].iteritems():
            if 'message' in value:
                click.echo(value['message'])

    return retval


@root.command('rename-node')
@click.argument('name')
@click.argument('new-name')
@click.pass_context
def rename_node(ctx, name, new_name):
    """
    Change the name of a node.

    NAME must be the name of a node that you control. NEW_NAME is the
    desired new name. It must adhere to the same naming rules as for
    the create-node command, namely, it must begin with a letter and
    consist of only lowercase letters, numbers, and hyphen.
    """
    client = ControllerClient()
    node = client.find_node(name)
    if node is None:
        click.echo("Node {} was not found.".format(name))
        return
    node['name'] = new_name
    result = client.save_node(node)
    click.echo(util.format_result(result))
    return result



import click

from .comm import pdserver_request


@click.group()
@click.pass_context
def server(ctx):
    """
    Commands to work with the ParaDrop server (controller).
    """
    ctx.obj['routers_url'] = ctx.obj['pdserver_url'] + "/api/routers"


@server.command()
@click.pass_context
def list_routers(ctx):
    """
    List routers.
    """
    url = ctx.obj['routers_url']
    result = pdserver_request('GET', url)
    routers = result.json()

    for router in routers:
        print("{} {} {}".format(router['_id'], router['name'], router['online']))


@server.command()
@click.pass_context
@click.argument('token')
def claim(ctx, token):
    """
    Claim an existing router.
    """
    url = ctx.obj['routers_url'] + '/claim'
    data = {
        'claim_token': token
    }
    result = pdserver_request('POST', url, json=data)
    if result.ok:
        router = result.json()
        print("Claimed router: {}".format(router['name']))
    else:
        print("There was an error claiming the router.")
        print("Please check that your claim token is correct.")


@routers.command()
@click.pass_context
@click.argument('name')
@click.option('--orphaned/--not-orphaned', default=False)
@click.option('--claim', default=None)
def create_router(ctx, name, orphaned, claim):
    """
    Create a new router.
    """
    url = ctx.obj['routers_url']
    data = {
        'name': name,
        'orphaned': orphaned
    }
    if claim is not None:
        data['claim_token'] = claim
    pdserver_request('POST', url, json=data)


@server.command()
@click.pass_context
@click.argument('router_id')
def delete_router(ctx, router_id):
    """
    Delete a router.
    """
    url = ctx.obj['routers_url'] + "/" + router_id
    pdserver_request('DELETE', url)



from setuptools import setup, find_packages

setup(
    name="pdtools",
    version='0.11.0',
    author="ParaDrop Labs",
    description="ParaDrop development tools",
    install_requires=[
        'appdirs~=1.4.3',
        'arrow~=0.10.0',
        'click~=6.7',
        'future~=0.16.0',
        'GitPython~=2.1.5',
        'PyYAML~=3.12',
        'requests~=2.18.1',
        'six~=1.10.0',
        'websocket-client~=0.40.0'
    ],

    packages=find_packages(),

    entry_points={
        'console_scripts': [
            'pdtools = pdtools.__main__:main',
        ],
    },
)



import glob
import os
import subprocess

import appdirs
import click

from . import cloud
from . import node


# Bash script content to enable tab completion.
# This was generated by running `_PDTOOLS_COMPLETE=source pdtools`.
PDTOOLS_COMPLETE = """
_pdtools_completion() {
    COMPREPLY=( $( env COMP_WORDS="${COMP_WORDS[*]}" \
                   COMP_CWORD=$COMP_CWORD \
                   _PDTOOLS_COMPLETE=complete $1 ) )
    return 0
}

complete -F _pdtools_completion -o default pdtools;
"""


@click.command('wizard')
@click.pass_context
def root(ctx):
    """
    Set up environment for development.
    """
    click.echo("This wizard will help set up your development environment "
               "for programming with Paradrop.\n")

    config_dir = appdirs.user_config_dir('pdtools', 'paradrop')
    env_path = os.path.join(config_dir, 'env.sh')

    new_account = False
    have_account = click.confirm("Do you have a developer account on paradrop.org")
    if not have_account:
        if click.confirm("Would you like to register one now"):
            click.echo("")

            # Keep trying until the registration succeeds.
            result = None
            while result is None:
                result = ctx.invoke(cloud.register)
                click.echo("")

            # Set flags - user has an account, and it is new.
            have_account = True
            new_account = True

    if have_account:
        click.echo("Let's log in to your developer account. This will give you "
                   "the ability to manage compute nodes remotely and publish "
                   "chutes to the store.\n")

        # Warning for new accounts - we need them to verify their email.
        if new_account:
            click.echo("Important: Since this is a new developer account, "
                       "please make sure "
                       "you have followed the instructions in the email to "
                       "activate your account before proceeding to log in.\n")

        # Keep trying until login succeeds.
        result = False
        while not result:
            result = ctx.invoke(cloud.login)
            click.echo("")

        click.echo("Checking for nodes that are listed on your account...\n")
        ctx.invoke(cloud.list_nodes)
        click.echo("")

        while click.confirm("Do you have a claim token for a node not listed above"):
            token = click.prompt("Claim token").strip()
            ctx.invoke(cloud.claim_node, token=token)
            click.echo("")

    click.echo("If you have network access to a node, you can use the "
               "pdtools node commands for development and testing on "
               "the node.")
    have_node = click.confirm("Do you have a node to configure")
    ready = False
    while have_node and not ready:
        target = click.prompt("Node address").strip()
        click.echo("")

        click.echo("Let's log in to the node. This will give you direct "
                   "access to install and configure chutes on the node. "
                   "Please note that the node username/password is different "
                   "from your paradrop.org account.\n")

        ctx.obj['target'] = target
        ready = ctx.invoke(node.login)
        click.echo("")

    if have_node:
        click.echo("Next, we'll look for SSH public keys in order to enable "
                   "SSH access to your node.\n")

        search_path = os.path.expanduser("~/.ssh/*.pub")
        pubkeys = glob.glob(search_path)
        if len(pubkeys) == 0:
            click.echo("No keys were found in ~/.ssh/. We can generate one now "
                       "or you can do this later by running ssh-keygen "
                       "and adding the key with pdtools node import-ssh-key <path>.")
            if click.confirm("Generate an SSH key now"):
                cmd = ["ssh-keygen", "-b", "4096"]
                subprocess.call(cmd)

        for path in glob.iglob(search_path):
            if click.confirm("Use the public key {}".format(path)):
                ctx.invoke(node.import_ssh_key, path=path)

    with open(env_path, "w") as output:
        output.write("# This file was automatically generated by pdtools wizard.\n\n")
        output.write(PDTOOLS_COMPLETE)

    click.echo("")
    click.echo("Setup complete!\n")

    click.echo("If you would like to enable tab completion for pdtools "
               "run the following command next.\n")
    click.echo("  source {}\n".format(env_path))

    if have_node:
        click.echo("To run pdtools node commands on your target node without typing "
                   "the address every time, you can set the following variable.\n")
        click.echo("  export PDTOOLS_NODE_TARGET={}\n".format(target))

    click.echo("These changes will only apply to your current terminal session. "
               "To make the changes persistent, add these commands to your .bashrc "
               "or equivalent configuration file.")
